from datetime import datetime
from math import isfinite
from typing import TYPE_CHECKING, Annotated, Optional

import strawberry
from strawberry.relay import Node, NodeID
from strawberry.scalars import JSON
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.context import Context

from .Annotation import Annotation
from .AnnotationSource import AnnotationSource
from .AnnotatorKind import AnnotatorKind

if TYPE_CHECKING:
    from .Span import Span
    from .User import User


@strawberry.type
class DocumentAnnotation(Node, Annotation):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.DocumentAnnotation]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("DocumentAnnotation ID mismatch")

    @strawberry.field(description="Name of the annotation, e.g. 'helpfulness' or 'relevance'.")  # type: ignore
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.name),
            )
        return val

    @strawberry.field(description="The kind of annotator that produced the annotation.")  # type: ignore
    async def annotator_kind(
        self,
        info: Info[Context, None],
    ) -> AnnotatorKind:
        if self.db_record:
            val = self.db_record.annotator_kind
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.annotator_kind),
            )
        return AnnotatorKind(val)

    @strawberry.field(
        description="Value of the annotation in the form of a string, e.g. "
        "'helpful' or 'not helpful'. Note that the label is not necessarily binary."
    )  # type: ignore
    async def label(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.label
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.label),
            )
        return val

    @strawberry.field(
        description="Value of the annotation in the form of a numeric score.",
    )  # type: ignore
    async def score(
        self,
        info: Info[Context, None],
    ) -> Optional[float]:
        if self.db_record:
            val = self.db_record.score
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.score),
            )
        return val if val is not None and isfinite(val) else None

    @strawberry.field(
        description="The annotator's explanation for the annotation result (i.e. "
        "score or label, or both) given to the subject."
    )  # type: ignore
    async def explanation(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.explanation
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.explanation),
            )
        return val

    @strawberry.field(description="The metadata associated with the annotation.")  # type: ignore
    async def metadata(
        self,
        info: Info[Context, None],
    ) -> JSON:
        if self.db_record:
            val = self.db_record.metadata_
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.metadata_),
            )
        return val

    @strawberry.field(description="The position of the annotation in the document.")  # type: ignore
    async def document_position(
        self,
        info: Info[Context, None],
    ) -> int:
        if self.db_record:
            val = self.db_record.document_position
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.document_position),
            )
        return val

    @strawberry.field(description="The identifier of the annotation.")  # type: ignore
    async def identifier(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.identifier
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.identifier),
            )
        return val

    @strawberry.field(description="The source of the annotation.")  # type: ignore
    async def source(
        self,
        info: Info[Context, None],
    ) -> AnnotationSource:
        if self.db_record:
            val = self.db_record.source
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.source),
            )
        return AnnotationSource(val)

    @strawberry.field(description="The date and time when the annotation was created.")  # type: ignore
    async def created_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.created_at
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.created_at),
            )
        return val

    @strawberry.field(description="The date and time when the annotation was last updated.")  # type: ignore
    async def updated_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.updated_at
        else:
            val = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.updated_at),
            )
        return val

    @strawberry.field(description="The span associated with the annotation.")  # type: ignore
    async def span(
        self,
        info: Info[Context, None],
    ) -> Annotated["Span", strawberry.lazy(".Span")]:
        if self.db_record:
            span_rowid = self.db_record.span_rowid
        else:
            span_rowid = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.span_rowid),
            )
        from .Span import Span

        return Span(id=span_rowid)

    @strawberry.field(description="The user that produced the annotation.")  # type: ignore
    async def user(
        self,
        info: Info[Context, None],
    ) -> Optional[Annotated["User", strawberry.lazy(".User")]]:
        if self.db_record:
            user_id = self.db_record.user_id
        else:
            user_id = await info.context.data_loaders.document_annotation_fields.load(
                (self.id, models.DocumentAnnotation.user_id),
            )
        if user_id is None:
            return None
        from .User import User

        return User(id=user_id)



import strawberry
from sqlalchemy import delete, select
from sqlalchemy.exc import IntegrityError as PostgreSQLIntegrityError
from sqlalchemy.orm import load_only
from sqlean.dbapi2 import IntegrityError as SQLiteIntegrityError  # type: ignore[import-untyped]
from strawberry.relay import GlobalID
from strawberry.types import Info

from phoenix.config import DEFAULT_PROJECT_NAME
from phoenix.db import models
from phoenix.server.api.auth import IsNotReadOnly, IsNotViewer
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest, Conflict
from phoenix.server.api.input_types.ClearProjectInput import ClearProjectInput
from phoenix.server.api.input_types.CreateProjectInput import CreateProjectInput
from phoenix.server.api.queries import Query
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.api.types.Project import Project, to_gql_project
from phoenix.server.dml_event import ProjectDeleteEvent, ProjectInsertEvent, SpanDeleteEvent


@strawberry.type
class ProjectMutationPayload:
    project: Project
    query: Query


@strawberry.type
class ProjectMutationMixin:
    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def create_project(
        self,
        info: Info[Context, None],
        input: CreateProjectInput,
    ) -> ProjectMutationPayload:
        if not (name := input.name.strip()):
            raise BadRequest("Name cannot be empty")
        description = (input.description or "").strip() or None
        gradient_start_color = (input.gradient_start_color or "").strip() or None
        gradient_end_color = (input.gradient_end_color or "").strip() or None
        project = models.Project(
            name=name,
            description=description,
            gradient_start_color=gradient_start_color,
            gradient_end_color=gradient_end_color,
        )
        try:
            async with info.context.db() as session:
                session.add(project)
        except (PostgreSQLIntegrityError, SQLiteIntegrityError):
            raise Conflict(f"Project with name '{name}' already exists")
        info.context.event_queue.put(ProjectInsertEvent((project.id,)))
        return ProjectMutationPayload(project=to_gql_project(project), query=Query())

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def delete_project(self, info: Info[Context, None], id: GlobalID) -> Query:
        project_id = from_global_id_with_expected_type(global_id=id, expected_type_name="Project")
        async with info.context.db() as session:
            project = await session.scalar(
                select(models.Project)
                .where(models.Project.id == project_id)
                .options(load_only(models.Project.name))
            )
            if project is None:
                raise ValueError(f"Unknown project: {id}")
            if project.name == DEFAULT_PROJECT_NAME:
                raise ValueError(f"Cannot delete the {DEFAULT_PROJECT_NAME} project")
            await session.delete(project)
        info.context.event_queue.put(ProjectDeleteEvent((project_id,)))
        return Query()

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def clear_project(self, info: Info[Context, None], input: ClearProjectInput) -> Query:
        project_id = from_global_id_with_expected_type(
            global_id=input.id, expected_type_name="Project"
        )
        delete_statement = (
            delete(models.Trace)
            .where(models.Trace.project_rowid == project_id)
            .returning(models.Trace.project_session_rowid)
        )
        if input.end_time:
            delete_statement = delete_statement.where(models.Trace.start_time < input.end_time)
        async with info.context.db() as session:
            deleted_trace_project_session_ids = await session.scalars(delete_statement)
            session_ids_to_delete = [
                id_ for id_ in set(deleted_trace_project_session_ids) if id_ is not None
            ]
            # Process deletions in chunks of 10000 to avoid PostgreSQL argument limit
            chunk_size = 10000
            stmt = delete(models.ProjectSession)
            for i in range(0, len(session_ids_to_delete), chunk_size):
                chunk = session_ids_to_delete[i : i + chunk_size]
                await session.execute(stmt.where(models.ProjectSession.id.in_(chunk)))
        info.context.event_queue.put(SpanDeleteEvent((project_id,)))
        return Query()



"""Add auth_method column to users table and migrate existing authentication data.

This migration:
1. Adds a new 'auth_method' column to the users table that indicates whether a user
   authenticates via local password ('LOCAL') or external OAuth2 ('OAUTH2')
2. Migrates existing authentication data to populate the new column:
   - Sets 'LOCAL' for users with password_hash
   - Sets 'OAUTH2' for users with OAuth2 credentials
3. Adds appropriate constraints to ensure data integrity:
   - NOT NULL constraint on auth_method
   - 'valid_auth_method': ensures only 'LOCAL' or 'OAUTH2' values
   - 'local_auth_has_password_no_oauth': ensures LOCAL users have password credentials and
     do not have OAuth2 credentials
   - 'non_local_auth_has_no_password': ensures OAUTH2 users do not have password credentials
4. Removes legacy constraints that are replaced by the new column:
   - 'password_hash_and_salt': ensures password_hash and password_salt are consistent
   - 'exactly_one_auth_method': replaced by auth_method column and its constraints
   - 'oauth2_client_id_and_user_id': replaced by auth_method column and its constraints
5. Drops redundant single column indices:
   - 'ix_users_oauth2_client_id' and 'ix_users_oauth2_user_id' are removed as they are
     redundant with the unique constraint 'uq_users_oauth2_client_id_oauth2_user_id',
     which already provides the necessary composite index for lookups

The migration uses batch_alter_table to ensure compatibility with both SQLite and PostgreSQL.
This approach allows us to:
- Add the column as nullable initially
- Update the values based on existing authentication data
- Make the column NOT NULL after populating
- Add appropriate constraints
- Remove legacy constraints
- Drop redundant indices

The downgrade path:
1. Recreates the legacy constraints:
   - 'password_hash_and_salt': ensures password_hash and password_salt are consistent
   - 'exactly_one_auth_method': ensures exactly one auth method is set
   - 'oauth2_client_id_and_user_id': ensures OAuth2 credentials are consistent
2. Removes the auth_method column and its associated constraints
3. Recreates the single column indices to maintain backward compatibility:
   - 'ix_users_oauth2_client_id'
   - 'ix_users_oauth2_user_id'

Revision ID: 6a88424799fe
Revises: 8a3764fe7f1a
Create Date: 2025-05-01 08:08:22.700715

"""  # noqa: E501

from typing import Sequence, Union

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision: str = "6a88424799fe"
down_revision: Union[str, None] = "8a3764fe7f1a"
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    """Upgrade the database schema to include the auth_method column.

    This function:
    1. Adds the auth_method column as nullable
    2. Populates the column based on existing authentication data:
       - 'LOCAL' for users with password_hash
       - 'OAUTH2' for users with OAuth2 credentials
    3. Makes the column NOT NULL after populating
    4. Adds CHECK constraints to ensure data integrity:
       - 'valid_auth_method': ensures only 'LOCAL' or 'OAUTH2' values
       - 'local_auth_has_password_no_oauth': ensures LOCAL users have password credentials and
          do not have OAuth2 credentials
       - 'non_local_auth_has_no_password': ensures OAUTH2 users do not have password credentials
    5. Removes legacy constraints that are replaced by the new column:
       - 'password_hash_and_salt'
       - 'exactly_one_auth_method'
       - 'oauth2_client_id_and_user_id'
    6. Drops redundant single column indices:
       - 'ix_users_oauth2_client_id' and 'ix_users_oauth2_user_id' are removed as they are
         redundant with the unique constraint 'uq_users_oauth2_client_id_oauth2_user_id',
         which already provides the necessary composite index for lookups

    The implementation uses batch_alter_table for compatibility with both
    SQLite and PostgreSQL databases.

    Raises:
        sqlalchemy.exc.SQLAlchemyError: If database operations fail
    """  # noqa: E501
    with op.batch_alter_table("users") as batch_op:
        # For SQLite, first add the column as nullable
        batch_op.add_column(sa.Column("auth_method", sa.String, nullable=True))

    with op.batch_alter_table("users") as batch_op:
        batch_op.execute("""
            UPDATE users
            SET auth_method = CASE
            WHEN password_hash IS NOT NULL THEN 'LOCAL' ELSE 'OAUTH2' END
        """)
        # Make the column non-nullable
        batch_op.alter_column("auth_method", nullable=False, existing_nullable=True)

        # Drop both old constraints as they're now redundant
        batch_op.drop_constraint("password_hash_and_salt", type_="check")
        batch_op.drop_constraint("exactly_one_auth_method", type_="check")
        batch_op.drop_constraint("oauth2_client_id_and_user_id", type_="check")

        # Drop redundant single column indices, because a composite index already
        # exists in the uniqueness constraint for (client_id, user_id)
        batch_op.drop_index("ix_users_oauth2_client_id")
        batch_op.drop_index("ix_users_oauth2_user_id")

        # Add CHECK constraint to ensure only valid values are allowed
        batch_op.create_check_constraint(
            "valid_auth_method",
            "auth_method IN ('LOCAL', 'OAUTH2')",
        )
        batch_op.create_check_constraint(
            "local_auth_has_password_no_oauth",
            "auth_method != 'LOCAL' "
            "OR (password_hash IS NOT NULL AND password_salt IS NOT NULL "
            "AND oauth2_client_id IS NULL AND oauth2_user_id IS NULL)",
        )
        batch_op.create_check_constraint(
            "non_local_auth_has_no_password",
            "auth_method = 'LOCAL' OR (password_hash IS NULL AND password_salt IS NULL)",
        )


def downgrade() -> None:
    """Downgrade the database schema by removing the auth_method column.

    This function:
    1. Recreates the legacy constraints that were removed in the upgrade:
       - 'password_hash_and_salt': ensures password_hash and password_salt are consistent
       - 'exactly_one_auth_method': ensures exactly one auth method is set
       - 'oauth2_client_id_and_user_id': ensures OAuth2 credentials are consistent
    2. Removes the auth_method column and its associated CHECK constraints:
       - 'non_local_auth_has_no_password'
       - 'local_auth_has_password_no_oauth'
       - 'valid_auth_method'
    3. Recreates the single column indices to maintain backward compatibility:
       - 'ix_users_oauth2_client_id'
       - 'ix_users_oauth2_user_id'

    The implementation uses batch_alter_table to ensure compatibility with both
    SQLite and PostgreSQL databases.

    Raises:
        sqlalchemy.exc.SQLAlchemyError: If database operations fail
    """  # noqa: E501
    # Use batch_alter_table for SQLite compatibility
    # This ensures the downgrade works on both SQLite and PostgreSQL
    with op.batch_alter_table("users") as batch_op:
        # Drop the CHECK constraint and column
        batch_op.drop_constraint("non_local_auth_has_no_password", type_="check")
        batch_op.drop_constraint("local_auth_has_password_no_oauth", type_="check")
        batch_op.drop_constraint("valid_auth_method", type_="check")

        # Recreate single column indices
        batch_op.create_index("ix_users_oauth2_user_id", ["oauth2_user_id"])
        batch_op.create_index("ix_users_oauth2_client_id", ["oauth2_client_id"])

        # Recreate both old constraints that were dropped in upgrade
        batch_op.create_check_constraint(
            "oauth2_client_id_and_user_id",
            "(oauth2_client_id IS NULL) = (oauth2_user_id IS NULL)",
        )
        batch_op.create_check_constraint(
            "exactly_one_auth_method",
            "(password_hash IS NULL) != (oauth2_client_id IS NULL)",
        )
        batch_op.create_check_constraint(
            "password_hash_and_salt",
            "(password_hash IS NULL) = (password_salt IS NULL)",
        )

        # Remove added column
        batch_op.drop_column("auth_method")



from typing import Optional, Union

import strawberry
from strawberry import UNSET
from strawberry.relay import GlobalID
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.types.AnnotationSource import AnnotationSource
from phoenix.server.api.types.node import from_global_id_with_expected_type


@strawberry.input
class AnnotationFilterCondition:
    names: Optional[list[str]] = UNSET
    sources: Optional[list[AnnotationSource]] = UNSET
    user_ids: Optional[list[Optional[GlobalID]]] = UNSET

    def __post_init__(self) -> None:
        if isinstance(self.names, list) and not self.names:
            raise BadRequest("names must be a non-empty list")
        if isinstance(self.sources, list) and not self.sources:
            raise BadRequest("sources must be a non-empty list")
        if isinstance(self.user_ids, list) and not self.user_ids:
            raise BadRequest("user ids must be a non-empty list")


@strawberry.input
class AnnotationFilter:
    include: Optional[AnnotationFilterCondition] = UNSET
    exclude: Optional[AnnotationFilterCondition] = UNSET

    def __post_init__(self) -> None:
        if self.include is UNSET and self.exclude is UNSET:
            raise BadRequest("include and exclude cannot both be unset")


_Annotation: TypeAlias = Union[
    models.SpanAnnotation,
    models.TraceAnnotation,
    models.ProjectSessionAnnotation,
]


def satisfies_filter(annotation: _Annotation, filter: AnnotationFilter) -> bool:
    """
    Returns true if the annotation satisfies the filter and false otherwise.
    """
    annotation_source = AnnotationSource(annotation.source)
    if include := filter.include:
        if include.names and annotation.name not in include.names:
            return False
        if include.sources and annotation_source not in include.sources:
            return False
        if include.user_ids:
            user_rowids = [
                from_global_id_with_expected_type(user_id, "User") if user_id is not None else None
                for user_id in include.user_ids
            ]
            if annotation.user_id not in user_rowids:
                return False
    if exclude := filter.exclude:
        if exclude.names and annotation.name in exclude.names:
            return False
        if exclude.sources and annotation_source in exclude.sources:
            return False
        if exclude.user_ids:
            user_rowids = [
                from_global_id_with_expected_type(user_id, "User") if user_id is not None else None
                for user_id in exclude.user_ids
            ]
            if annotation.user_id in user_rowids:
                return False
    return True



import strawberry

from phoenix.server.api.types.CostBreakdown import CostBreakdown


@strawberry.type
class SpanCostDetailSummaryEntry:
    token_type: str
    is_prompt: bool
    value: CostBreakdown = strawberry.field(default_factory=CostBreakdown)



# This file is generated. Do not edit by hand.
# ruff: noqa: E501

from ._models import ClassificationEvaluatorConfig, PromptMessage

TOOL_INVOCATION_CLASSIFICATION_EVALUATOR_CONFIG = ClassificationEvaluatorConfig(
    name="tool_invocation",
    description="For determining if a tool was invoked correctly with proper arguments, formatting, and safe content. Requires conversation context, available tool schemas, and the LLM's tool invocation(s).",
    optimization_direction="maximize",
    messages=[
        PromptMessage(
            role="user",
            content='You are an impartial judge evaluating an LLM\'s tool-calling behavior, specifically whether the LLM invoked a tool (or tools) correctly with valid arguments and proper formatting.\nYour task: Determine whether the LLM\'s tool invocation(s) were correct or incorrect based on: - The full conversation context (including all previous turns, not just the most recent message) - The available tool schemas - The LLM\'s tool invocation(s) with arguments\nIMPORTANT - Tool Invocation vs. Tool Selection: - You are ONLY evaluating the tool invocation, not the tool selection. - If the tool selection is incorrect or not relevant to the user\'s query, but the tool invocation is correct, return "correct". - If the tool selection is correct but the tool invocation is incorrect, return "incorrect".\nIMPORTANT - Multi-Tool Invocations: - The LLM may invoke MULTIPLE tools in a single response. This is valid and expected for complex requests. - When multiple tools are invoked, evaluate EACH tool invocation independently. - Return "correct" only if ALL tool invocations are correct. - Return "incorrect" if ANY tool invocation has an error.\nIMPORTANT - Conversation Context: - Read the entire conversation history carefully, not just the final user message. - Argument values may need to be extracted from EARLIER turns in the conversation (e.g., user mentions a location, date, or quantity in a previous message). - The LLM should use context from the full conversation to populate argument values correctly.\nCriteria Return "correct" only when ALL of the following are true for EVERY tool invocation: - JSON is properly structured (if applicable). - All required fields/parameters are present. - No hallucinated or nonexistent fields (all fields exist in the tool schema). - Argument values match the user\'s intent from the conversation context (correct types, realistic values). - No unsafe content (e.g., PII like SSNs, credit card numbers, passwords) in arguments.\nReturn "incorrect" if ANY of the following are true for ANY tool invocation: - The invocation contains hallucinated or nonexistent fields not in the schema. - Required fields/parameters are missing. - JSON is improperly formatted or malformed. - Argument values are incorrect, hallucinated, or do not match user intent from the conversation. - Arguments contain unsafe content (e.g., PII, sensitive data that should not be passed).\nBefore providing your final judgment, explain your reasoning and consider: - How many tools were invoked? Evaluate each one. - Does each tool invocation match the schema for that tool? - Are all required parameters provided with appropriate values for each invocation? - Are there any extra fields that don\'t exist in the schema? - Looking at the FULL conversation: do the argument values accurately reflect what the user requested across all messages? - Is there any unsafe or sensitive content in any of the arguments? - Check that you are not evaluating the tool selection, only the tool invocation.\n<data> <context> {{input}} </context>\n<available_tools> {{available_tools}} </available_tools>\n<tool_invocation> {{tool_selection}} </tool_invocation> </data>\nGiven the above data, is the tool invocation correct or incorrect?',
        )
    ],
    choices={"correct": 1.0, "incorrect": 0.0},
)



import asyncio
import logging
from asyncio import Queue, as_completed
from collections import deque
from dataclasses import dataclass, field
from functools import singledispatchmethod
from time import perf_counter, time
from typing import Any, AsyncIterator, Awaitable, Callable, Iterable, Optional, cast

from openinference.semconv.trace import SpanAttributes
from typing_extensions import TypeAlias

import phoenix.trace.v1 as pb
from phoenix.db import models
from phoenix.db.insertion.constants import DEFAULT_RETRY_ALLOWANCE, DEFAULT_RETRY_DELAY_SEC
from phoenix.db.insertion.document_annotation import DocumentAnnotationQueueInserter
from phoenix.db.insertion.evaluation import (
    InsertEvaluationError,
    insert_evaluation,
)
from phoenix.db.insertion.helpers import (
    DataManipulation,
    DataManipulationEvent,
    should_calculate_span_cost,
)
from phoenix.db.insertion.session_annotation import SessionAnnotationQueueInserter
from phoenix.db.insertion.span import SpanInsertionEvent, insert_span
from phoenix.db.insertion.span_annotation import SpanAnnotationQueueInserter
from phoenix.db.insertion.trace_annotation import TraceAnnotationQueueInserter
from phoenix.db.insertion.types import Insertables, Precursors
from phoenix.server.daemons.span_cost_calculator import (
    SpanCostCalculator,
)
from phoenix.server.dml_event import DmlEvent, SpanInsertEvent
from phoenix.server.prometheus import (
    BULK_LOADER_EVALUATION_INSERTIONS,
    BULK_LOADER_EXCEPTIONS,
    BULK_LOADER_LAST_ACTIVITY,
    BULK_LOADER_SPAN_EXCEPTIONS,
    BULK_LOADER_SPAN_INSERTION_TIME,
    SPAN_QUEUE_SIZE,
)
from phoenix.server.types import CanPutItem, DbSessionFactory
from phoenix.trace.schemas import Span

logger = logging.getLogger(__name__)

ProjectRowId: TypeAlias = int
ProjectName: TypeAlias = str


@dataclass(frozen=True)
class TransactionResult:
    updated_project_rowids: set[ProjectRowId] = field(default_factory=set)


class BulkInserter:
    def __init__(
        self,
        db: DbSessionFactory,
        *,
        event_queue: CanPutItem[DmlEvent],
        span_cost_calculator: SpanCostCalculator,
        initial_batch_of_spans: Iterable[tuple[Span, ProjectName]] = (),
        initial_batch_of_evaluations: Iterable[pb.Evaluation] = (),
        sleep: float = 0.1,
        max_ops_per_transaction: int = 1000,
        max_queue_size: int = 1000,
        max_spans_queue_size: Optional[int] = None,
        retry_delay_sec: float = DEFAULT_RETRY_DELAY_SEC,
        retry_allowance: int = DEFAULT_RETRY_ALLOWANCE,
    ) -> None:
        """
        :param db: A function to initiate a new database session.
        :param initial_batch_of_spans: Initial batch of spans to insert.
        :param sleep: The time to sleep between bulk insertions
        :param max_ops_per_transaction: The maximum number of operations to dequeue from
        the operations queue for each transaction.
        :param max_queue_size: The maximum length of the operations queue.
        """
        self._db = db
        self._running = False
        self._sleep = sleep
        self._max_ops_per_transaction = max_ops_per_transaction
        self._operations: Optional[Queue[DataManipulation]] = None
        self._max_queue_size = max_queue_size
        self._max_spans_queue_size = max_spans_queue_size
        self._spans: deque[tuple[Span, ProjectName]] = deque(initial_batch_of_spans)
        self._evaluations: deque[pb.Evaluation] = deque(initial_batch_of_evaluations)
        self._task: Optional[asyncio.Task[None]] = None
        self._event_queue = event_queue
        self._retry_delay_sec = retry_delay_sec
        self._retry_allowance = retry_allowance
        self._queue_inserters = _QueueInserters(db, self._retry_delay_sec, self._retry_allowance)
        self._span_cost_calculator = span_cost_calculator

    @property
    def is_full(self) -> bool:
        return bool(self._max_spans_queue_size and self._max_spans_queue_size <= len(self._spans))

    async def __aenter__(
        self,
    ) -> tuple[
        Callable[[Any], Awaitable[None]],
        Callable[[Span, str], Awaitable[None]],
        Callable[[pb.Evaluation], Awaitable[None]],
        Callable[[DataManipulation], None],
    ]:
        self._running = True
        self._operations = Queue(maxsize=self._max_queue_size)
        self._task = asyncio.create_task(self._bulk_insert())
        return (
            self._enqueue_annotations,
            self._enqueue_span,
            self._enqueue_evaluation,
            self._enqueue_operation,
        )

    async def __aexit__(self, *args: Any) -> None:
        self._running = False
        if self._task:
            self._task.cancel()
            self._task = None

    async def _enqueue_annotations(self, *items: Any) -> None:
        await self._queue_inserters.enqueue(*items)

    def _enqueue_operation(self, operation: DataManipulation) -> None:
        cast("Queue[DataManipulation]", self._operations).put_nowait(operation)

    async def _enqueue_span(self, span: Span, project_name: str) -> None:
        self._spans.append((span, project_name))

    async def _enqueue_evaluation(self, evaluation: pb.Evaluation) -> None:
        self._evaluations.append(evaluation)

    async def _process_events(self, events: Iterable[Optional[DataManipulationEvent]]) -> None: ...

    async def _bulk_insert(self) -> None:
        assert isinstance(self._operations, Queue)
        # start first insert immediately if the inserter has not run recently
        while (
            self._running
            or not self._queue_inserters.empty
            or not self._operations.empty()
            or self._spans
            or self._evaluations
        ):
            BULK_LOADER_LAST_ACTIVITY.set(time())
            SPAN_QUEUE_SIZE.set(len(self._spans))
            if (
                self._queue_inserters.empty
                and self._operations.empty()
                and not self._spans
                and not self._evaluations
            ):
                await asyncio.sleep(self._sleep)
                continue
            ops_remaining = self._max_ops_per_transaction
            async with self._db() as session:
                while ops_remaining and not self._operations.empty():
                    ops_remaining -= 1
                    op = await self._operations.get()
                    try:
                        async with session.begin_nested():
                            await op(session)
                    except Exception as e:
                        BULK_LOADER_EXCEPTIONS.inc()
                        logger.exception(str(e))
            # It's important to grab the buffers at the same time so there's
            # no race condition, since an eval insertion will fail if the span
            # it references doesn't exist. Grabbing the eval buffer later may
            # include an eval whose span is in the queue but missed being
            # included in the span buffer that was grabbed previously.
            num_spans_to_insert = min(self._max_ops_per_transaction, len(self._spans))
            num_evals_to_insert = min(self._max_ops_per_transaction, len(self._evaluations))
            # Spans should be inserted before the evaluations, since an evaluation
            # insertion will fail if the span it references doesn't exist.
            await self._insert_spans(num_spans_to_insert)
            await self._insert_evaluations(num_evals_to_insert)
            async for event in self._queue_inserters.insert():
                self._event_queue.put(event)
            await asyncio.sleep(self._sleep)

    async def _insert_spans(self, num_spans_to_insert: int) -> None:
        if not num_spans_to_insert or not self._spans:
            return
        project_ids = set()
        span_costs: list[models.SpanCost] = []
        try:
            start = perf_counter()
            async with self._db() as session:
                while num_spans_to_insert > 0:
                    num_spans_to_insert -= 1
                    if not self._spans:
                        break
                    span, project_name = self._spans.popleft()
                    result: Optional[SpanInsertionEvent] = None
                    try:
                        async with session.begin_nested():
                            result = await insert_span(session, span, project_name)
                    except Exception:
                        BULK_LOADER_SPAN_EXCEPTIONS.inc()
                        logger.exception(
                            f"Failed to insert span with span_id={span.context.span_id}"
                        )
                    if result is None:
                        continue
                    project_ids.add(result.project_rowid)
                    try:
                        if not should_calculate_span_cost(span.attributes):
                            continue
                        span_cost = self._span_cost_calculator.calculate_cost(
                            span.start_time,
                            span.attributes,
                        )
                    except Exception:
                        logger.exception(
                            f"Failed to calculate span cost for span with "
                            f"span_id={span.context.span_id}"
                        )
                    else:
                        if span_cost is None:
                            continue
                        span_cost.span_rowid = result.span_rowid
                        span_cost.trace_rowid = result.trace_rowid
                        span_costs.append(span_cost)
            BULK_LOADER_SPAN_INSERTION_TIME.observe(perf_counter() - start)
        except Exception:
            BULK_LOADER_SPAN_EXCEPTIONS.inc()
            logger.exception("Failed to insert spans")
        if project_ids:
            self._event_queue.put(SpanInsertEvent(tuple(project_ids)))
        if not span_costs:
            return
        try:
            async with self._db() as session:
                session.add_all(span_costs)
        except Exception:
            logger.exception("Failed to insert span costs")

    async def _insert_evaluations(self, num_evals_to_insert: int) -> None:
        if not num_evals_to_insert or not self._evaluations:
            return
        try:
            async with self._db() as session:
                while num_evals_to_insert > 0:
                    num_evals_to_insert -= 1
                    if not self._evaluations:
                        break
                    evaluation = self._evaluations.popleft()
                    BULK_LOADER_EVALUATION_INSERTIONS.inc()
                    try:
                        async with session.begin_nested():
                            await insert_evaluation(session, evaluation)
                    except InsertEvaluationError as error:
                        BULK_LOADER_EXCEPTIONS.inc()
                        logger.exception(f"Failed to insert evaluation: {str(error)}")
        except Exception:
            BULK_LOADER_EXCEPTIONS.inc()
            logger.exception("Failed to insert evaluations")


class _QueueInserters:
    def __init__(
        self,
        db: DbSessionFactory,
        retry_delay_sec: float = DEFAULT_RETRY_DELAY_SEC,
        retry_allowance: int = DEFAULT_RETRY_ALLOWANCE,
    ) -> None:
        self._db = db
        args = (db, retry_delay_sec, retry_allowance)
        self._span_annotations = SpanAnnotationQueueInserter(*args)
        self._trace_annotations = TraceAnnotationQueueInserter(*args)
        self._document_annotations = DocumentAnnotationQueueInserter(*args)
        self._session_annotations = SessionAnnotationQueueInserter(*args)
        self._queues = (
            self._span_annotations,
            self._trace_annotations,
            self._document_annotations,
            self._session_annotations,
        )

    async def insert(self) -> AsyncIterator[DmlEvent]:
        if self.empty:
            return
        for coro in as_completed([q.insert() for q in self._queues if not q.empty]):
            if events := cast(Optional[list[DmlEvent]], await coro):
                for event in events:
                    yield event

    @property
    def empty(self) -> bool:
        return all(q.empty for q in self._queues)

    async def enqueue(self, *items: Any) -> None:
        for item in items:
            await self._enqueue(item)

    @singledispatchmethod
    async def _enqueue(self, item: Any) -> None: ...

    @_enqueue.register(Precursors.SpanAnnotation)
    @_enqueue.register(Insertables.SpanAnnotation)
    async def _(self, item: Precursors.SpanAnnotation) -> None:
        await self._span_annotations.enqueue(item)

    @_enqueue.register(Precursors.TraceAnnotation)
    @_enqueue.register(Insertables.TraceAnnotation)
    async def _(self, item: Precursors.TraceAnnotation) -> None:
        await self._trace_annotations.enqueue(item)

    @_enqueue.register(Precursors.DocumentAnnotation)
    @_enqueue.register(Insertables.DocumentAnnotation)
    async def _(self, item: Precursors.DocumentAnnotation) -> None:
        await self._document_annotations.enqueue(item)

    @_enqueue.register(Precursors.SessionAnnotation)
    @_enqueue.register(Insertables.SessionAnnotation)
    async def _(self, item: Precursors.SessionAnnotation) -> None:
        await self._session_annotations.enqueue(item)


LLM_MODEL_NAME = SpanAttributes.LLM_MODEL_NAME
LLM_PROVIDER = SpanAttributes.LLM_PROVIDER
LLM_TOKEN_COUNT_COMPLETION = SpanAttributes.LLM_TOKEN_COUNT_COMPLETION
LLM_TOKEN_COUNT_COMPLETION_DETAILS_AUDIO = SpanAttributes.LLM_TOKEN_COUNT_COMPLETION_DETAILS_AUDIO
LLM_TOKEN_COUNT_COMPLETION_DETAILS_REASONING = (
    SpanAttributes.LLM_TOKEN_COUNT_COMPLETION_DETAILS_REASONING
)
LLM_TOKEN_COUNT_PROMPT = SpanAttributes.LLM_TOKEN_COUNT_PROMPT
LLM_TOKEN_COUNT_PROMPT_DETAILS_AUDIO = SpanAttributes.LLM_TOKEN_COUNT_PROMPT_DETAILS_AUDIO
LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_READ = SpanAttributes.LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_READ
LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_WRITE = (
    SpanAttributes.LLM_TOKEN_COUNT_PROMPT_DETAILS_CACHE_WRITE
)



from typing import TYPE_CHECKING, Any, Callable, Optional, Sequence, Union

from phoenix.server.api.types.GenerativeProvider import GenerativeProviderKey

if TYPE_CHECKING:
    from phoenix.server.api.helpers.playground_clients import PlaygroundStreamingClient

ModelName = Union[str, None]
ModelKey = tuple[GenerativeProviderKey, ModelName]

PROVIDER_DEFAULT = None


class SingletonMeta(type):
    _instances: dict[Any, Any] = dict()

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class PlaygroundClientRegistry(metaclass=SingletonMeta):
    def __init__(self) -> None:
        self._registry: dict[
            GenerativeProviderKey, dict[ModelName, Optional[type["PlaygroundStreamingClient"]]]
        ] = {}

    def get_client(
        self,
        provider_key: GenerativeProviderKey,
        model_name: ModelName,
    ) -> Optional[type["PlaygroundStreamingClient"]]:
        provider_registry = self._registry.get(provider_key, {})
        client_class = provider_registry.get(model_name)
        if client_class is None and None in provider_registry:
            client_class = provider_registry[PROVIDER_DEFAULT]  # Fallback to provider default
        return client_class

    def list_all_providers(
        self,
    ) -> list[GenerativeProviderKey]:
        return [provider_key for provider_key in self._registry]

    def list_models(self, provider_key: GenerativeProviderKey) -> list[str]:
        provider_registry = self._registry.get(provider_key, {})
        return [model_name for model_name in provider_registry.keys() if model_name is not None]

    def list_all_models(self) -> list[ModelKey]:
        return [
            (provider_key, model_name)
            for provider_key, provider_registry in self._registry.items()
            for model_name in provider_registry.keys()
        ]


PLAYGROUND_CLIENT_REGISTRY: PlaygroundClientRegistry = PlaygroundClientRegistry()


def register_llm_client(
    provider_key: GenerativeProviderKey,
    model_names: Sequence[ModelName],
) -> Callable[[type["PlaygroundStreamingClient"]], type["PlaygroundStreamingClient"]]:
    def decorator(cls: type["PlaygroundStreamingClient"]) -> type["PlaygroundStreamingClient"]:
        provider_registry = PLAYGROUND_CLIENT_REGISTRY._registry.setdefault(provider_key, {})
        for model_name in model_names:
            provider_registry[model_name] = cls
        return cls

    return decorator



from enum import Enum

import strawberry


@strawberry.enum
class TimeBinScale(Enum):
    MINUTE = "minute"
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    YEAR = "year"


@strawberry.input
class TimeBinConfig:
    scale: TimeBinScale = strawberry.field(
        default=TimeBinScale.HOUR, description="The scale of time bins for aggregation."
    )
    utc_offset_minutes: int = strawberry.field(
        default=0, description="Offset in minutes from UTC for local time binning."
    )



from dataclasses import dataclass
from typing import Optional

from sqlalchemy import func, select, tuple_
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

ExperimentID: TypeAlias = int
DatasetExampleID: TypeAlias = int
AnnotationName: TypeAlias = str
MeanAnnotationScore: TypeAlias = float


@dataclass
class AnnotationSummary:
    annotation_name: AnnotationName
    mean_score: Optional[MeanAnnotationScore]


Key: TypeAlias = tuple[ExperimentID, DatasetExampleID]
Result: TypeAlias = list[AnnotationSummary]


class ExperimentRepeatedRunGroupAnnotationSummariesDataLoader(DataLoader[Key, Result]):
    def __init__(
        self,
        db: DbSessionFactory,
    ) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        annotation_summaries_query = (
            select(
                models.ExperimentRun.experiment_id.label("experiment_id"),
                models.ExperimentRun.dataset_example_id.label("dataset_example_id"),
                models.ExperimentRunAnnotation.name.label("annotation_name"),
                func.avg(models.ExperimentRunAnnotation.score).label("mean_score"),
            )
            .select_from(models.ExperimentRunAnnotation)
            .join(
                models.ExperimentRun,
                models.ExperimentRunAnnotation.experiment_run_id == models.ExperimentRun.id,
            )
            .where(
                tuple_(
                    models.ExperimentRun.experiment_id, models.ExperimentRun.dataset_example_id
                ).in_(set(keys))
            )
            .group_by(
                models.ExperimentRun.experiment_id,
                models.ExperimentRun.dataset_example_id,
                models.ExperimentRunAnnotation.name,
            )
        )
        async with self._db() as session:
            annotation_summaries = (await session.execute(annotation_summaries_query)).all()
        annotation_summaries_by_key: dict[Key, list[AnnotationSummary]] = {}
        for summary in annotation_summaries:
            key = (summary.experiment_id, summary.dataset_example_id)
            gql_summary = AnnotationSummary(
                annotation_name=summary.annotation_name,
                mean_score=summary.mean_score,
            )
            if key not in annotation_summaries_by_key:
                annotation_summaries_by_key[key] = []
            annotation_summaries_by_key[key].append(gql_summary)
        return [
            sorted(
                annotation_summaries_by_key.get(key, []),
                key=lambda summary: summary.annotation_name,
            )
            for key in keys
        ]



from strawberry.relay import GlobalID


def get_user_identifier(user_id: int) -> str:
    """
    Generates an annotation identifier unique to the user.
    """
    user_gid = str(GlobalID(type_name="User", node_id=str(user_id)))
    return f"px-app:{user_gid}"



import re
from abc import ABC, abstractmethod
from collections.abc import Iterable
from string import Formatter
from typing import Any


class TemplateFormatter(ABC):
    @abstractmethod
    def parse(self, template: str) -> set[str]:
        """
        Parse the template and return a set of variable names.
        """
        raise NotImplementedError

    def format(self, template: str, **variables: Any) -> str:
        """
        Formats the template with the given variables.
        """
        template_variable_names = self.parse(template)
        if missing_template_variables := template_variable_names - set(variables.keys()):
            raise TemplateFormatterError(
                f"Missing template variable(s): {', '.join(missing_template_variables)}"
            )
        return self._format(template, template_variable_names, **variables)

    @abstractmethod
    def _format(self, template: str, variable_names: Iterable[str], **variables: Any) -> str:
        raise NotImplementedError


class NoOpFormatter(TemplateFormatter):
    """
    No-op template formatter.

    Examples:

    >>> formatter = NoOpFormatter()
    >>> formatter.format("hello")
    'hello'
    """

    def parse(self, template: str) -> set[str]:
        return set()

    def _format(self, template: str, *args: Any, **variables: Any) -> str:
        return template


class FStringTemplateFormatter(TemplateFormatter):
    """
    Regular f-string template formatter.

    Examples:

    >>> formatter = FStringTemplateFormatter()
    >>> formatter.format("{hello}", hello="world")
    'world'
    """

    def parse(self, template: str) -> set[str]:
        return set(field_name for _, field_name, _, _ in Formatter().parse(template) if field_name)

    def _format(self, template: str, variable_names: Iterable[str], **variables: Any) -> str:
        return template.format(**variables)


class MustacheTemplateFormatter(TemplateFormatter):
    """
    Mustache template formatter.

    Examples:

    >>> formatter = MustacheTemplateFormatter()
    >>> formatter.format("{{ hello }}", hello="world")
    'world'
    """

    PATTERN = re.compile(r"(?<!\\){{\s*(\w+)\s*}}")

    def parse(self, template: str) -> set[str]:
        return set(match for match in re.findall(self.PATTERN, template))

    def _format(self, template: str, variable_names: Iterable[str], **variables: Any) -> str:
        for variable_name in variable_names:
            template = re.sub(
                pattern=rf"(?<!\\){{{{\s*{variable_name}\s*}}}}",
                repl=str(variables[variable_name]),
                string=template,
            )
        return template


class TemplateFormatterError(Exception):
    """
    An error raised when template formatting fails.
    """

    pass



from typing import TYPE_CHECKING, Any, Awaitable, Callable, Iterable, Optional

import grpc
from grpc.aio import RpcContext, Server, ServerInterceptor
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceRequest,
    ExportTraceServiceResponse,
)
from opentelemetry.proto.collector.trace.v1.trace_service_pb2_grpc import (
    TraceServiceServicer,
    add_TraceServiceServicer_to_server,
)
from starlette.concurrency import run_in_threadpool
from typing_extensions import TypeAlias

from phoenix.auth import CanReadToken
from phoenix.config import (
    TLSConfigVerifyClient,
    get_env_grpc_port,
    get_env_tls_config,
    get_env_tls_enabled_for_grpc,
)
from phoenix.server.bearer_auth import ApiKeyInterceptor
from phoenix.trace.otel import decode_otlp_span
from phoenix.trace.schemas import Span
from phoenix.utilities.project import get_project_name

if TYPE_CHECKING:
    from opentelemetry.trace import TracerProvider

ProjectName: TypeAlias = str


class Servicer(TraceServiceServicer):  # type: ignore[misc,unused-ignore]
    def __init__(
        self,
        enqueue_span: Callable[[Span, ProjectName], Awaitable[None]],
    ) -> None:
        super().__init__()
        self._enqueue_span = enqueue_span

    async def Export(
        self,
        request: ExportTraceServiceRequest,
        context: RpcContext,
    ) -> ExportTraceServiceResponse:
        for resource_spans in request.resource_spans:
            project_name = get_project_name(resource_spans.resource.attributes)
            for scope_span in resource_spans.scope_spans:
                for otlp_span in scope_span.spans:
                    span = await run_in_threadpool(decode_otlp_span, otlp_span)
                    await self._enqueue_span(span, project_name)
        return ExportTraceServiceResponse()


class GrpcServer:
    def __init__(
        self,
        enqueue_span: Callable[[Span, ProjectName], Awaitable[None]],
        tracer_provider: Optional["TracerProvider"] = None,
        enable_prometheus: bool = False,
        disabled: bool = False,
        token_store: Optional[CanReadToken] = None,
        interceptors: Iterable[ServerInterceptor] = (),
    ) -> None:
        self._enqueue_span = enqueue_span
        self._server: Optional[Server] = None
        self._tracer_provider = tracer_provider
        self._enable_prometheus = enable_prometheus
        self._disabled = disabled
        self._token_store = token_store
        self._interceptors = list(interceptors)

    async def __aenter__(self) -> None:
        interceptors = self._interceptors
        if self._disabled:
            return
        if self._token_store:
            interceptors.append(ApiKeyInterceptor(self._token_store))
        if self._enable_prometheus:
            ...
            # TODO: convert to async interceptor
            # from py_grpc_prometheus.prometheus_server_interceptor import PromServerInterceptor
            #
            # interceptors.append(PromServerInterceptor())
        if self._tracer_provider is not None:
            from opentelemetry.instrumentation.grpc import GrpcAioInstrumentorServer

            GrpcAioInstrumentorServer().instrument(tracer_provider=self._tracer_provider)  # type: ignore
        server = grpc.aio.server(
            options=(("grpc.so_reuseport", 0),),
            interceptors=interceptors,
        )
        if get_env_tls_enabled_for_grpc():
            assert (tls_config := get_env_tls_config())
            private_key_certificate_chain_pairs = [(tls_config.key_data, tls_config.cert_data)]
            server_credentials = (
                grpc.ssl_server_credentials(
                    private_key_certificate_chain_pairs,
                    root_certificates=tls_config.ca_data,
                    require_client_auth=True,
                )
                if isinstance(tls_config, TLSConfigVerifyClient)
                else grpc.ssl_server_credentials(private_key_certificate_chain_pairs)
            )
            server.add_secure_port(f"[::]:{get_env_grpc_port()}", server_credentials)
        else:
            server.add_insecure_port(f"[::]:{get_env_grpc_port()}")
        add_TraceServiceServicer_to_server(Servicer(self._enqueue_span), server)  # type: ignore[no-untyped-call,unused-ignore]
        await server.start()
        self._server = server

    async def __aexit__(self, *args: Any, **kwargs: Any) -> None:
        if self._server is None:
            return
        await self._server.stop(5)
        self._server = None
        if self._tracer_provider is not None:
            from opentelemetry.instrumentation.grpc import GrpcAioInstrumentorServer

            GrpcAioInstrumentorServer().uninstrument()  # type: ignore



from typing import Optional

from fastapi import APIRouter, Depends, HTTPException, Path, Query
from pydantic import Field
from sqlalchemy import select
from starlette.requests import Request
from strawberry.relay import GlobalID

from phoenix.config import DEFAULT_PROJECT_NAME
from phoenix.db import models
from phoenix.db.helpers import exclude_experiment_projects
from phoenix.server.api.routers.v1.models import V1RoutesBaseModel
from phoenix.server.api.routers.v1.utils import (
    PaginatedResponseBody,
    ResponseBody,
    _get_project_by_identifier,
    add_errors_to_responses,
)
from phoenix.server.api.types.Project import Project as ProjectNodeType
from phoenix.server.authorization import is_not_locked, require_admin

router = APIRouter(tags=["projects"])


class ProjectData(V1RoutesBaseModel):
    name: str = Field(..., min_length=1)
    description: Optional[str] = None


class Project(ProjectData):
    id: str


class GetProjectsResponseBody(PaginatedResponseBody[Project]):
    pass


class GetProjectResponseBody(ResponseBody[Project]):
    pass


class CreateProjectRequestBody(ProjectData):
    pass


class CreateProjectResponseBody(ResponseBody[Project]):
    pass


class UpdateProjectRequestBody(V1RoutesBaseModel):
    description: Optional[str] = None


class UpdateProjectResponseBody(ResponseBody[Project]):
    pass


@router.get(
    "/projects",
    operation_id="getProjects",
    summary="List all projects",  # noqa: E501
    description="Retrieve a paginated list of all projects in the system.",  # noqa: E501
    response_description="A list of projects with pagination information",  # noqa: E501
    responses=add_errors_to_responses(
        [
            422,
        ]
    ),
)
async def get_projects(
    request: Request,
    cursor: Optional[str] = Query(
        default=None,
        description="Cursor for pagination (project ID)",
    ),
    limit: int = Query(
        default=100, description="The max number of projects to return at a time.", gt=0
    ),
    include_experiment_projects: bool = Query(
        default=False,
        description="Include experiment projects in the response. Experiment projects are created from running experiments.",  # noqa: E501
    ),
) -> GetProjectsResponseBody:
    """
    Retrieve a paginated list of all projects in the system.

    Args:
        request (Request): The FastAPI request object.
        cursor (Optional[str]): Pagination cursor (project ID).
        limit (int): Maximum number of projects to return per request.
        include_experiment_projects (bool): Flag to include experiment projects in the response.
            Experiment projects are created from running experiments.

    Returns:
        GetProjectsResponseBody: Response containing a list of projects and pagination information.

    Raises:
        HTTPException: If the cursor format is invalid.
    """  # noqa: E501
    stmt = select(models.Project).order_by(models.Project.id.desc())
    if not include_experiment_projects:
        stmt = exclude_experiment_projects(stmt)
    async with request.app.state.db() as session:
        if cursor:
            try:
                cursor_id = GlobalID.from_id(cursor).node_id
                stmt = stmt.filter(models.Project.id <= int(cursor_id))
            except ValueError:
                raise HTTPException(
                    detail=f"Invalid cursor format: {cursor}",
                    status_code=422,
                )

        stmt = stmt.limit(limit + 1)
        projects = (await session.scalars(stmt)).all()

        if not projects:
            return GetProjectsResponseBody(next_cursor=None, data=[])

        next_cursor = None
        if len(projects) == limit + 1:
            last_project = projects[-1]
            next_cursor = str(GlobalID(ProjectNodeType.__name__, str(last_project.id)))
            projects = projects[:-1]

        project_responses = [_to_project_response(project) for project in projects]
    return GetProjectsResponseBody(next_cursor=next_cursor, data=project_responses)


@router.get(
    "/projects/{project_identifier}",
    operation_id="getProject",
    summary="Get project by ID or name",  # noqa: E501
    description="Retrieve a specific project using its unique identifier: either project ID or project name. Note: When using a project name as the identifier, it cannot contain slash (/), question mark (?), or pound sign (#) characters.",  # noqa: E501
    response_description="The requested project",  # noqa: E501
    responses=add_errors_to_responses(
        [
            404,
            422,
        ]
    ),
)
async def get_project(
    request: Request,
    project_identifier: str = Path(
        description="The project identifier: either project ID or project name. If using a project name, it cannot contain slash (/), question mark (?), or pound sign (#) characters.",  # noqa: E501
    ),
) -> GetProjectResponseBody:
    """
    Retrieve a specific project by its ID or name.

    Args:
        request (Request): The FastAPI request object.
        project_identifier (str): The project identifier: either project ID or project name.
            If using a project name, it cannot contain slash (/), question mark (?), or pound sign (#) characters.

    Returns:
        GetProjectResponseBody: Response containing the requested project.

    Raises:
        HTTPException: If the project identifier format is invalid or the project is not found.
    """  # noqa: E501
    async with request.app.state.db() as session:
        project = await _get_project_by_identifier(session, project_identifier)
    data = _to_project_response(project)
    return GetProjectResponseBody(data=data)


@router.post(
    "/projects",
    dependencies=[Depends(is_not_locked)],
    operation_id="createProject",
    summary="Create a new project",  # noqa: E501
    description="Create a new project with the specified configuration.",  # noqa: E501
    response_description="The newly created project",  # noqa: E501
    responses=add_errors_to_responses(
        [
            422,
        ]
    ),
)
async def create_project(
    request: Request,
    request_body: CreateProjectRequestBody,
) -> CreateProjectResponseBody:
    """
    Create a new project.

    Args:
        request (Request): The FastAPI request object.
        request_body (CreateProjectRequestBody): The request body containing project data.

    Returns:
        CreateProjectResponseBody: Response containing the created project.

    Raises:
        HTTPException: If any validation error occurs.
    """
    async with request.app.state.db() as session:
        project = models.Project(
            name=request_body.name,
            description=request_body.description,
        )
        session.add(project)
        await session.flush()
    data = _to_project_response(project)
    return CreateProjectResponseBody(data=data)


@router.put(
    "/projects/{project_identifier}",
    dependencies=[Depends(require_admin), Depends(is_not_locked)],
    operation_id="updateProject",
    summary="Update a project by ID or name",  # noqa: E501
    description="Update an existing project with new configuration. Project names cannot be changed. The project identifier is either project ID or project name. Note: When using a project name as the identifier, it cannot contain slash (/), question mark (?), or pound sign (#) characters.",  # noqa: E501
    response_description="The updated project",  # noqa: E501
    responses=add_errors_to_responses(
        [
            403,
            404,
            422,
        ]
    ),
)
async def update_project(
    request: Request,
    request_body: UpdateProjectRequestBody,
    project_identifier: str = Path(
        description="The project identifier: either project ID or project name. If using a project name, it cannot contain slash (/), question mark (?), or pound sign (#) characters.",  # noqa: E501
    ),
) -> UpdateProjectResponseBody:
    """
    Update an existing project.

    Args:
        request (Request): The FastAPI request object.
        request_body (UpdateProjectRequestBody): The request body containing the new description.
        project_identifier (str): The project identifier: either project ID or project name.
            If using a project name, it cannot contain slash (/), question mark (?), or pound sign (#) characters.

    Returns:
        UpdateProjectResponseBody: Response containing the updated project.

    Raises:
        HTTPException: If the project identifier format is invalid or the project is not found.
    """  # noqa: E501
    async with request.app.state.db() as session:
        project = await _get_project_by_identifier(session, project_identifier)

        # Update the description if provided
        if request_body.description is not None:
            project.description = request_body.description

    data = _to_project_response(project)
    return UpdateProjectResponseBody(data=data)


@router.delete(
    "/projects/{project_identifier}",
    dependencies=[Depends(require_admin)],
    operation_id="deleteProject",
    summary="Delete a project by ID or name",  # noqa: E501
    description="Delete an existing project and all its associated data. The project identifier is either project ID or project name. The default project cannot be deleted. Note: When using a project name as the identifier, it cannot contain slash (/), question mark (?), or pound sign (#) characters.",  # noqa: E501
    response_description="No content returned on successful deletion",  # noqa: E501
    status_code=204,
    responses=add_errors_to_responses(
        [
            403,
            404,
            422,
        ]
    ),
)
async def delete_project(
    request: Request,
    project_identifier: str = Path(
        description="The project identifier: either project ID or project name. If using a project name, it cannot contain slash (/), question mark (?), or pound sign (#) characters.",  # noqa: E501
    ),
) -> None:
    """
    Delete an existing project.

    Args:
        request (Request): The FastAPI request object.
        project_identifier (str): The project identifier: either project ID or project name.
            If using a project name, it cannot contain slash (/), question mark (?), or pound sign (#) characters.

    Returns:
        None: Returns a 204 No Content response on success.

    Raises:
        HTTPException: If the project identifier format is invalid, the project is not found, or it's the default project.
    """  # noqa: E501
    async with request.app.state.db() as session:
        project = await _get_project_by_identifier(session, project_identifier)

        # The default project must not be deleted - it's forbidden
        if project.name == DEFAULT_PROJECT_NAME:
            raise HTTPException(
                status_code=403,
                detail="The default project cannot be deleted",
            )

        await session.delete(project)
    return None


def _to_project_response(project: models.Project) -> Project:
    return Project(
        id=str(GlobalID(ProjectNodeType.__name__, str(project.id))),
        name=project.name,
        description=project.description,
    )



from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

DatasetID: TypeAlias = int
Key: TypeAlias = DatasetID
Result: TypeAlias = list[models.DatasetLabel]


class DatasetLabelsDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        dataset_ids = keys
        async with self._db() as session:
            labels: dict[Key, Result] = {}
            for dataset_id, label in await session.execute(
                select(models.DatasetsDatasetLabel.dataset_id, models.DatasetLabel)
                .select_from(models.DatasetLabel)
                .join(
                    models.DatasetsDatasetLabel,
                    models.DatasetLabel.id == models.DatasetsDatasetLabel.dataset_label_id,
                )
                .where(models.DatasetsDatasetLabel.dataset_id.in_(dataset_ids))
            ):
                if dataset_id not in labels:
                    labels[dataset_id] = []
                labels[dataset_id].append(label)
        return [
            sorted(labels.get(dataset_id, []), key=lambda label: label.name) for dataset_id in keys
        ]



from __future__ import annotations

from collections import defaultdict
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Optional, Union

import pandas as pd
import strawberry
from aioitertools.itertools import islice
from openinference.semconv.trace import SpanAttributes
from sqlalchemy import desc, or_, select
from strawberry import ID, UNSET, lazy
from strawberry.relay import Connection, GlobalID, Node, NodeID
from strawberry.types import Info
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.api.context import Context
from phoenix.server.api.input_types.AnnotationFilter import AnnotationFilter, satisfies_filter
from phoenix.server.api.input_types.TraceAnnotationSort import TraceAnnotationSort
from phoenix.server.api.types.AnnotationSummary import AnnotationSummary
from phoenix.server.api.types.CostBreakdown import CostBreakdown
from phoenix.server.api.types.pagination import (
    Cursor,
    CursorString,
    connection_from_cursors_and_nodes,
)
from phoenix.server.api.types.SortDir import SortDir
from phoenix.server.api.types.Span import Span
from phoenix.server.api.types.SpanCostDetailSummaryEntry import SpanCostDetailSummaryEntry
from phoenix.server.api.types.SpanCostSummary import SpanCostSummary
from phoenix.server.api.types.TraceAnnotation import TraceAnnotation

if TYPE_CHECKING:
    from phoenix.server.api.types.Project import Project
    from phoenix.server.api.types.ProjectSession import ProjectSession

ProjectRowId: TypeAlias = int
TraceRowId: TypeAlias = int


@strawberry.type
class Trace(Node):
    id: NodeID[TraceRowId]
    db_record: strawberry.Private[Optional[models.Trace]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("Trace ID mismatch")

    @strawberry.field
    async def trace_id(
        self,
        info: Info[Context, None],
    ) -> ID:
        if self.db_record:
            trace_id = self.db_record.trace_id
        else:
            trace_id = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.trace_id),
            )
        return ID(trace_id)

    @strawberry.field
    async def start_time(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            start_time = self.db_record.start_time
        else:
            start_time = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.start_time),
            )
        return start_time

    @strawberry.field
    async def end_time(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            end_time = self.db_record.end_time
        else:
            end_time = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.end_time),
            )
        return end_time

    @strawberry.field
    async def latency_ms(
        self,
        info: Info[Context, None],
    ) -> Optional[float]:
        if self.db_record:
            latency_ms = self.db_record.latency_ms
        else:
            latency_ms = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.latency_ms),
            )
        return latency_ms

    @strawberry.field
    async def project(
        self,
        info: Info[Context, None],
    ) -> Annotated["Project", strawberry.lazy(".Project")]:
        if self.db_record:
            project_rowid = self.db_record.project_rowid
        else:
            project_rowid = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.project_rowid),
            )
        from phoenix.server.api.types.Project import Project

        return Project(id=project_rowid)

    @strawberry.field
    async def project_id(
        self,
        info: Info[Context, None],
    ) -> GlobalID:
        if self.db_record:
            project_rowid = self.db_record.project_rowid
        else:
            project_rowid = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.project_rowid),
            )
        from phoenix.server.api.types.Project import Project

        return GlobalID(type_name=Project.__name__, node_id=str(project_rowid))

    @strawberry.field
    async def project_session_id(
        self,
        info: Info[Context, None],
    ) -> Optional[GlobalID]:
        if self.db_record:
            project_session_rowid = self.db_record.project_session_rowid
        else:
            project_session_rowid = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.project_session_rowid),
            )
        if project_session_rowid is None:
            return None
        from phoenix.server.api.types.ProjectSession import ProjectSession

        return GlobalID(type_name=ProjectSession.__name__, node_id=str(project_session_rowid))

    @strawberry.field
    async def session(
        self,
        info: Info[Context, None],
    ) -> Union[Annotated["ProjectSession", lazy(".ProjectSession")], None]:
        if self.db_record:
            project_session_rowid = self.db_record.project_session_rowid
        else:
            project_session_rowid = await info.context.data_loaders.trace_fields.load(
                (self.id, models.Trace.project_session_rowid),
            )
        if project_session_rowid is None:
            return None

        stmt = select(models.ProjectSession).filter_by(id=project_session_rowid)
        async with info.context.db() as session:
            project_session = await session.scalar(stmt)
        if project_session is None:
            return None
        from .ProjectSession import ProjectSession

        return ProjectSession(id=project_session.id, db_record=project_session)

    @strawberry.field
    async def root_span(
        self,
        info: Info[Context, None],
    ) -> Optional[Span]:
        span_rowid = await info.context.data_loaders.trace_root_spans.load(self.id)
        if span_rowid is None:
            return None
        return Span(id=span_rowid)

    @strawberry.field
    async def num_spans(
        self,
        info: Info[Context, None],
    ) -> int:
        return await info.context.data_loaders.num_spans_per_trace.load(self.id)

    @strawberry.field
    async def spans(
        self,
        info: Info[Context, None],
        first: Optional[int] = 50,
        last: Optional[int] = UNSET,
        after: Optional[CursorString] = UNSET,
        before: Optional[CursorString] = UNSET,
        root_spans_only: Optional[bool] = UNSET,
        orphan_span_as_root_span: Optional[bool] = True,
    ) -> Connection[Span]:
        # Validate pagination arguments
        if isinstance(first, int) and first <= 0:
            raise ValueError('Argument "first" must be a positive int')

        # Build base query for spans in this trace
        base_query = (
            select(models.Span.id)
            .join(models.Trace)
            .where(models.Trace.id == self.id)
            # Sort descending because the root span tends to show up later
            # in the ingestion process.
            .order_by(desc(models.Span.id))
        )
        # Handle cursor pagination (forward pagination only)
        if after is not UNSET and after is not None:
            # Type narrowing: after is guaranteed to be str at this point
            assert after is not None  # Type narrowing for mypy
            try:
                cursor = Cursor.from_string(after)
            except Exception as e:
                raise ValueError(f"Invalid cursor format: {after}") from e
            # For descending order, "after" means we want spans with smaller IDs
            # (going forward in descending order)
            base_query = base_query.where(models.Span.id < cursor.rowid)
        # Note: backward pagination (last/before) is not yet implemented
        # as it requires more complex handling with reversed ordering
        if before is not UNSET or (last is not UNSET and last is not None):
            raise ValueError("Backward pagination (last/before) is not yet supported")

        # Build final query based on filtering requirements
        if root_spans_only:
            if orphan_span_as_root_span:
                # A root span is either a span with no parent_id or an orphan span
                # (a span whose parent_id references a span that doesn't exist in the current trace)
                # We need parent_id to check for orphan spans, so add it to the query
                # and create a CTE
                candidate_spans = base_query.add_columns(models.Span.parent_id).cte(
                    "candidate_spans"
                )
                # Subquery to get all span_ids that exist in this trace
                parent_spans_in_trace = (
                    select(models.Span.span_id)
                    .where(models.Span.trace_rowid == self.id)
                    .alias("parent_spans")
                )
                # Filter candidates to only root spans (NULL parent_id or orphan spans)
                stmt = (
                    select(candidate_spans.c.id)
                    .where(
                        or_(
                            candidate_spans.c.parent_id.is_(None),
                            ~select(1)
                            .where(candidate_spans.c.parent_id == parent_spans_in_trace.c.span_id)
                            .exists(),
                        )
                    )
                    .order_by(desc(candidate_spans.c.id))
                )
            else:
                # Only include explicit root spans (spans with parent_id = NULL)
                stmt = base_query.where(models.Span.parent_id.is_(None))
        else:
            # Return all spans (no root span filtering)
            stmt = base_query

        # Over-fetch by one to determine whether there's a next page
        limit = first if isinstance(first, int) else 50
        stmt = stmt.limit(limit + 1)

        cursors_and_nodes = []
        async with info.context.db() as session:
            span_rowids = await session.stream_scalars(stmt)
            async for span_rowid in islice(span_rowids, limit):
                cursor = Cursor(rowid=span_rowid)
                cursors_and_nodes.append((cursor, Span(id=span_rowid)))
            has_next_page = True
            try:
                await span_rowids.__anext__()
            except StopAsyncIteration:
                has_next_page = False
        return connection_from_cursors_and_nodes(
            cursors_and_nodes,
            has_previous_page=False,
            has_next_page=has_next_page,
        )

    @strawberry.field(description="Annotations associated with the trace.")  # type: ignore
    async def trace_annotations(
        self,
        info: Info[Context, None],
        sort: Optional[TraceAnnotationSort] = None,
    ) -> list[TraceAnnotation]:
        async with info.context.db() as session:
            stmt = select(models.TraceAnnotation).filter_by(trace_rowid=self.id)
            if sort:
                sort_col = getattr(models.TraceAnnotation, sort.col.value)
                if sort.dir is SortDir.desc:
                    stmt = stmt.order_by(sort_col.desc(), models.TraceAnnotation.id.desc())
                else:
                    stmt = stmt.order_by(sort_col.asc(), models.TraceAnnotation.id.asc())
            else:
                stmt = stmt.order_by(models.TraceAnnotation.created_at.desc())
            annotations = await session.scalars(stmt)
        return [
            TraceAnnotation(id=annotation.id, db_record=annotation) for annotation in annotations
        ]

    @strawberry.field(description="Summarizes each annotation (by name) associated with the trace")  # type: ignore
    async def trace_annotation_summaries(
        self,
        info: Info[Context, None],
        filter: Optional[AnnotationFilter] = None,
    ) -> list[AnnotationSummary]:
        """
        Retrieves and summarizes annotations associated with this span.

        This method aggregates annotation data by name and label, calculating metrics
        such as count of occurrences and sum of scores. The results are organized
        into a structured format that can be easily converted to a DataFrame.

        Args:
            info: GraphQL context information
            filter: Optional filter to apply to annotations before processing

        Returns:
            A list of AnnotationSummary objects, each containing:
            - name: The name of the annotation
            - data: A list of dictionaries with label statistics
        """
        # Load all annotations for this span from the data loader
        annotations = await info.context.data_loaders.trace_annotations_by_trace.load(self.id)

        # Apply filter if provided to narrow down the annotations
        if filter:
            annotations = [
                annotation for annotation in annotations if satisfies_filter(annotation, filter)
            ]

        @dataclass
        class Metrics:
            record_count: int = 0
            label_count: int = 0
            score_sum: float = 0
            score_count: int = 0

        summaries: defaultdict[str, defaultdict[Optional[str], Metrics]] = defaultdict(
            lambda: defaultdict(Metrics)
        )
        for annotation in annotations:
            metrics = summaries[annotation.name][annotation.label]
            metrics.record_count += 1
            metrics.label_count += int(annotation.label is not None)
            metrics.score_sum += annotation.score or 0
            metrics.score_count += int(annotation.score is not None)

        result: list[AnnotationSummary] = []
        for name, label_metrics in summaries.items():
            rows = [{"label": label, **asdict(metrics)} for label, metrics in label_metrics.items()]
            result.append(AnnotationSummary(name=name, df=pd.DataFrame(rows), simple_avg=True))
        return result

    @strawberry.field
    async def cost_summary(
        self,
        info: Info[Context, None],
    ) -> SpanCostSummary:
        loader = info.context.data_loaders.span_cost_summary_by_trace
        summary = await loader.load(self.id)
        return SpanCostSummary(
            prompt=CostBreakdown(
                tokens=summary.prompt.tokens,
                cost=summary.prompt.cost,
            ),
            completion=CostBreakdown(
                tokens=summary.completion.tokens,
                cost=summary.completion.cost,
            ),
            total=CostBreakdown(
                tokens=summary.total.tokens,
                cost=summary.total.cost,
            ),
        )

    @strawberry.field
    async def cost_detail_summary_entries(
        self,
        info: Info[Context, None],
    ) -> list[SpanCostDetailSummaryEntry]:
        loader = info.context.data_loaders.span_cost_detail_summary_entries_by_trace
        entries = await loader.load(self.id)
        return [
            SpanCostDetailSummaryEntry(
                token_type=entry.token_type,
                is_prompt=entry.is_prompt,
                value=CostBreakdown(tokens=entry.value.tokens, cost=entry.value.cost),
            )
            for entry in entries
        ]


INPUT_VALUE = SpanAttributes.INPUT_VALUE.split(".")
OUTPUT_VALUE = SpanAttributes.OUTPUT_VALUE.split(".")



from enum import Enum


class ModelProvider(Enum):
    OPENAI = "OPENAI"
    AZURE_OPENAI = "AZURE_OPENAI"
    ANTHROPIC = "ANTHROPIC"
    GOOGLE = "GOOGLE"
    DEEPSEEK = "DEEPSEEK"
    XAI = "XAI"
    OLLAMA = "OLLAMA"
    AWS = "AWS"



from typing import Optional

import strawberry
from sqlalchemy import delete, insert, select
from starlette.requests import Request
from strawberry import UNSET, Info

from phoenix.db import models
from phoenix.server.api.auth import IsLocked, IsNotReadOnly, IsNotViewer
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest, NotFound, Unauthorized
from phoenix.server.api.helpers.annotations import get_user_identifier
from phoenix.server.api.input_types.CreateTraceAnnotationInput import CreateTraceAnnotationInput
from phoenix.server.api.input_types.DeleteAnnotationsInput import DeleteAnnotationsInput
from phoenix.server.api.input_types.PatchAnnotationInput import PatchAnnotationInput
from phoenix.server.api.queries import Query
from phoenix.server.api.types.AnnotationSource import AnnotationSource
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.api.types.TraceAnnotation import TraceAnnotation
from phoenix.server.bearer_auth import PhoenixUser
from phoenix.server.dml_event import TraceAnnotationDeleteEvent, TraceAnnotationInsertEvent


@strawberry.type
class TraceAnnotationMutationPayload:
    trace_annotations: list[TraceAnnotation]
    query: Query


@strawberry.type
class TraceAnnotationMutationMixin:
    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer, IsLocked])  # type: ignore
    async def create_trace_annotations(
        self, info: Info[Context, None], input: list[CreateTraceAnnotationInput]
    ) -> TraceAnnotationMutationPayload:
        if not input:
            raise BadRequest("No trace annotations provided.")

        assert isinstance(request := info.context.request, Request)
        user_id: Optional[int] = None
        if "user" in request.scope and isinstance((user := info.context.user), PhoenixUser):
            user_id = int(user.identity)

        processed_annotations_map: dict[int, models.TraceAnnotation] = {}

        trace_rowids = []
        for idx, annotation_input in enumerate(input):
            try:
                trace_rowid = from_global_id_with_expected_type(annotation_input.trace_id, "Trace")
            except ValueError:
                raise BadRequest(
                    f"Invalid trace ID for annotation at index {idx}: {annotation_input.trace_id}"
                )
            trace_rowids.append(trace_rowid)

        async with info.context.db() as session:
            for idx, (trace_rowid, annotation_input) in enumerate(zip(trace_rowids, input)):
                resolved_identifier = ""
                if isinstance(annotation_input.identifier, str):
                    resolved_identifier = annotation_input.identifier
                elif annotation_input.source == AnnotationSource.APP and user_id is not None:
                    resolved_identifier = get_user_identifier(user_id)
                values = {
                    "trace_rowid": trace_rowid,
                    "name": annotation_input.name,
                    "label": annotation_input.label,
                    "score": annotation_input.score,
                    "explanation": annotation_input.explanation,
                    "annotator_kind": annotation_input.annotator_kind.value,
                    "metadata_": annotation_input.metadata,
                    "identifier": resolved_identifier,
                    "source": annotation_input.source.value,
                    "user_id": user_id,
                }

                processed_annotation: Optional[models.TraceAnnotation] = None

                # Check if an annotation with this trace_rowid, name, and identifier already exists
                q = select(models.TraceAnnotation).where(
                    models.TraceAnnotation.trace_rowid == trace_rowid,
                    models.TraceAnnotation.name == annotation_input.name,
                    models.TraceAnnotation.identifier == resolved_identifier,
                )
                existing_annotation = await session.scalar(q)

                if existing_annotation:
                    # Update existing annotation
                    existing_annotation.name = values["name"]
                    existing_annotation.label = values["label"]
                    existing_annotation.score = values["score"]
                    existing_annotation.explanation = values["explanation"]
                    existing_annotation.metadata_ = values["metadata_"]
                    existing_annotation.annotator_kind = values["annotator_kind"]
                    existing_annotation.source = values["source"]
                    existing_annotation.user_id = values["user_id"]
                    session.add(existing_annotation)
                    processed_annotation = existing_annotation

                if processed_annotation is None:
                    stmt = insert(models.TraceAnnotation).values(**values)
                    stmt = stmt.returning(models.TraceAnnotation)
                    result = await session.scalars(stmt)
                    processed_annotation = result.one()

                processed_annotations_map[idx] = processed_annotation

            await session.commit()

        inserted_annotation_ids = tuple(anno.id for anno in processed_annotations_map.values())
        if inserted_annotation_ids:
            info.context.event_queue.put(TraceAnnotationInsertEvent(inserted_annotation_ids))

        returned_annotations = [
            TraceAnnotation(
                id=processed_annotations_map[i].id, db_record=processed_annotations_map[i]
            )
            for i in sorted(processed_annotations_map.keys())
        ]

        return TraceAnnotationMutationPayload(
            trace_annotations=returned_annotations,
            query=Query(),
        )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer, IsLocked])  # type: ignore
    async def patch_trace_annotations(
        self, info: Info[Context, None], input: list[PatchAnnotationInput]
    ) -> TraceAnnotationMutationPayload:
        if not input:
            raise BadRequest("No trace annotations provided.")

        assert isinstance(request := info.context.request, Request)
        user_id: Optional[int] = None
        if "user" in request.scope and isinstance((user := info.context.user), PhoenixUser):
            user_id = int(user.identity)

        patch_by_id = {}
        for patch in input:
            try:
                trace_annotation_id = from_global_id_with_expected_type(
                    patch.annotation_id, "TraceAnnotation"
                )
            except ValueError:
                raise BadRequest(f"Invalid trace annotation ID: {patch.annotation_id}")
            if trace_annotation_id in patch_by_id:
                raise BadRequest(f"Duplicate patch for trace annotation ID: {trace_annotation_id}")
            patch_by_id[trace_annotation_id] = patch

        async with info.context.db() as session:
            trace_annotations_by_id = {}
            for trace_annotation in await session.scalars(
                select(models.TraceAnnotation).where(
                    models.TraceAnnotation.id.in_(patch_by_id.keys())
                )
            ):
                if trace_annotation.user_id != user_id:
                    raise Unauthorized(
                        "At least one trace annotation is not associated with the current user."
                    )
                trace_annotations_by_id[trace_annotation.id] = trace_annotation

            missing_trace_annotation_ids = set(patch_by_id.keys()) - set(
                trace_annotations_by_id.keys()
            )
            if missing_trace_annotation_ids:
                raise NotFound(
                    f"Could not find trace annotations with IDs: {missing_trace_annotation_ids}"
                )

            for trace_annotation_id, patch in patch_by_id.items():
                trace_annotation = trace_annotations_by_id[trace_annotation_id]
                if patch.name:
                    trace_annotation.name = patch.name
                if patch.annotator_kind:
                    trace_annotation.annotator_kind = patch.annotator_kind.value
                if patch.label is not UNSET:
                    trace_annotation.label = patch.label
                if patch.score is not UNSET:
                    trace_annotation.score = patch.score
                if patch.explanation is not UNSET:
                    trace_annotation.explanation = patch.explanation
                if patch.metadata is not UNSET:
                    assert isinstance(patch.metadata, dict)
                    trace_annotation.metadata_ = patch.metadata
                if patch.identifier is not UNSET:
                    trace_annotation.identifier = patch.identifier or ""
                session.add(trace_annotation)
            await session.commit()

        patched_annotations = [
            TraceAnnotation(id=trace_annotation.id, db_record=trace_annotation)
            for trace_annotation in trace_annotations_by_id.values()
        ]
        info.context.event_queue.put(TraceAnnotationInsertEvent(tuple(patch_by_id.keys())))
        return TraceAnnotationMutationPayload(
            trace_annotations=patched_annotations,
            query=Query(),
        )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def delete_trace_annotations(
        self, info: Info[Context, None], input: DeleteAnnotationsInput
    ) -> TraceAnnotationMutationPayload:
        if not input.annotation_ids:
            raise BadRequest("No trace annotation IDs provided.")

        trace_annotation_ids: dict[int, None] = {}  # use dict to preserve order
        for annotation_gid in input.annotation_ids:
            try:
                annotation_id = from_global_id_with_expected_type(annotation_gid, "TraceAnnotation")
            except ValueError:
                raise BadRequest(f"Invalid trace annotation ID: {annotation_gid}")
            if annotation_id in trace_annotation_ids:
                raise BadRequest(f"Duplicate trace annotation ID: {annotation_id}")
            trace_annotation_ids[annotation_id] = None

        assert isinstance(request := info.context.request, Request)
        user_id: Optional[int] = None
        user_is_admin = False
        if "user" in request.scope and isinstance((user := info.context.user), PhoenixUser):
            user_id = int(user.identity)
            user_is_admin = user.is_admin

        async with info.context.db() as session:
            result = await session.scalars(
                delete(models.TraceAnnotation)
                .where(models.TraceAnnotation.id.in_(trace_annotation_ids.keys()))
                .returning(models.TraceAnnotation)
            )
            deleted_annotations_by_id = {annotation.id: annotation for annotation in result.all()}

            if not user_is_admin and any(
                annotation.user_id != user_id for annotation in deleted_annotations_by_id.values()
            ):
                await session.rollback()
                raise Unauthorized(
                    "At least one trace annotation is not associated with the current user "
                    "and the current user is not an admin."
                )

            missing_trace_annotation_ids = set(trace_annotation_ids.keys()) - set(
                deleted_annotations_by_id.keys()
            )
            if missing_trace_annotation_ids:
                raise NotFound(
                    f"Could not find trace annotations with IDs: {missing_trace_annotation_ids}"
                )

        deleted_gql_annotations = [
            TraceAnnotation(
                id=deleted_annotations_by_id[id].id, db_record=deleted_annotations_by_id[id]
            )
            for id in trace_annotation_ids
        ]
        info.context.event_queue.put(
            TraceAnnotationDeleteEvent(tuple(deleted_annotations_by_id.keys()))
        )
        return TraceAnnotationMutationPayload(
            trace_annotations=deleted_gql_annotations, query=Query()
        )



from typing import Optional

from sqlalchemy import select
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

SpanID: TypeAlias = int
Key: TypeAlias = SpanID
Result: TypeAlias = Optional[models.SpanCost]


class SpanCostsDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        span_ids = list(set(keys))
        async with self._db() as session:
            costs = {
                span_cost.span_rowid: span_cost
                async for span_cost in await session.stream_scalars(
                    select(models.SpanCost).where(models.SpanCost.span_rowid.in_(span_ids))
                )
            }
        return [costs.get(span_id) for span_id in keys]



from datetime import datetime, timezone
from typing import Optional

import strawberry


@strawberry.input
class TimeRange:
    start: Optional[datetime] = strawberry.field(
        default=None,
        description="The start of the time range",
    )
    end: Optional[datetime] = strawberry.field(
        default=None,
        description="The end of the time range. Right exclusive.",
    )

    def __post_init__(self) -> None:
        if self.start:
            self.start = self.start.astimezone(timezone.utc)
        if self.end:
            self.end = self.end.astimezone(timezone.utc)

    def is_valid(self) -> bool:
        return not self.start or not self.end or self.start < self.end



from datetime import datetime
from typing import Optional

import strawberry


@strawberry.interface
class ApiKey:
    @strawberry.field(description="Name of the API key.")  # type: ignore
    async def name(self) -> str:
        raise NotImplementedError

    @strawberry.field(description="Description of the API key.")  # type: ignore
    async def description(self) -> Optional[str]:
        raise NotImplementedError

    @strawberry.field(description="The date and time the API key was created.")  # type: ignore
    async def created_at(self) -> datetime:
        raise NotImplementedError

    @strawberry.field(description="The date and time the API key will expire.")  # type: ignore
    async def expires_at(self) -> Optional[datetime]:
        raise NotImplementedError



from enum import Enum

import strawberry

from phoenix.server.api.types.SortDir import SortDir


@strawberry.enum
class SpanAnnotationColumn(Enum):
    createdAt = "created_at"
    name = "name"


@strawberry.input(description="The sort key and direction for SpanAnnotation connections")
class SpanAnnotationSort:
    col: SpanAnnotationColumn
    dir: SortDir



import strawberry


@strawberry.type
class TokenUsage:
    prompt: float = 0
    completion: float = 0

    @strawberry.field
    async def total(self) -> float:
        return self.prompt + self.completion



from enum import Enum
from typing import Any, ClassVar, Optional, Union

import strawberry
from openinference.semconv.trace import OpenInferenceLLMProviderValues, SpanAttributes

from phoenix.config import getenv
from phoenix.trace.attributes import get_attribute_value


@strawberry.enum
class GenerativeProviderKey(Enum):
    OPENAI = "OpenAI"
    ANTHROPIC = "Anthropic"
    AZURE_OPENAI = "Azure OpenAI"
    GOOGLE = "Google Gemini"
    DEEPSEEK = "DeepSeek"
    XAI = "xAI"
    OLLAMA = "Ollama"
    AWS = "AWS Bedrock"


@strawberry.type
class GenerativeProviderCredentialConfig:
    env_var_name: str
    is_required: bool


@strawberry.type
class GenerativeProvider:
    name: str
    key: GenerativeProviderKey

    model_provider_to_model_prefix_map: ClassVar[dict[GenerativeProviderKey, list[str]]] = {
        GenerativeProviderKey.AZURE_OPENAI: [],
        GenerativeProviderKey.ANTHROPIC: ["claude"],
        GenerativeProviderKey.OPENAI: ["gpt", "o1"],
        GenerativeProviderKey.GOOGLE: ["gemini"],
        GenerativeProviderKey.DEEPSEEK: ["deepseek"],
        GenerativeProviderKey.XAI: ["grok"],
        GenerativeProviderKey.OLLAMA: ["llama", "mistral", "codellama", "phi", "qwen", "gemma"],
        GenerativeProviderKey.AWS: ["nova", "titan"],
    }

    attribute_provider_to_generative_provider_map: ClassVar[dict[str, GenerativeProviderKey]] = {
        OpenInferenceLLMProviderValues.OPENAI.value: GenerativeProviderKey.OPENAI,
        OpenInferenceLLMProviderValues.ANTHROPIC.value: GenerativeProviderKey.ANTHROPIC,
        OpenInferenceLLMProviderValues.AZURE.value: GenerativeProviderKey.AZURE_OPENAI,
        OpenInferenceLLMProviderValues.GOOGLE.value: GenerativeProviderKey.GOOGLE,
        OpenInferenceLLMProviderValues.AWS.value: GenerativeProviderKey.AWS,
        # Note: DeepSeek uses OpenAI compatibility but we can't duplicate the key in the dict
        # The provider will be determined through model name prefix matching instead
        # Note: xAI uses OpenAI compatibility but we can't duplicate the key in the dict
        # The provider will be determined through model name prefix matching instead
        # Note: Ollama uses OpenAI compatibility but we can't duplicate the key in the dict
        # The provider will be determined through model name prefix matching instead
    }

    """
    A map of model provider keys to their credential requirements.
    E.x. OpenAI requires a single API key
    """
    model_provider_to_credential_requirements_map: ClassVar[
        dict[GenerativeProviderKey, list[GenerativeProviderCredentialConfig]]
    ] = {
        GenerativeProviderKey.AZURE_OPENAI: [
            GenerativeProviderCredentialConfig(
                env_var_name="AZURE_OPENAI_API_KEY", is_required=True
            )
        ],
        GenerativeProviderKey.ANTHROPIC: [
            GenerativeProviderCredentialConfig(env_var_name="ANTHROPIC_API_KEY", is_required=True)
        ],
        GenerativeProviderKey.OPENAI: [
            GenerativeProviderCredentialConfig(env_var_name="OPENAI_API_KEY", is_required=True)
        ],
        GenerativeProviderKey.GOOGLE: [
            GenerativeProviderCredentialConfig(env_var_name="GEMINI_API_KEY", is_required=True)
        ],
        GenerativeProviderKey.DEEPSEEK: [
            GenerativeProviderCredentialConfig(env_var_name="DEEPSEEK_API_KEY", is_required=True)
        ],
        GenerativeProviderKey.XAI: [
            GenerativeProviderCredentialConfig(env_var_name="XAI_API_KEY", is_required=True)
        ],
        GenerativeProviderKey.OLLAMA: [],
        GenerativeProviderKey.AWS: [
            GenerativeProviderCredentialConfig(env_var_name="AWS_ACCESS_KEY_ID", is_required=True),
            GenerativeProviderCredentialConfig(
                env_var_name="AWS_SECRET_ACCESS_KEY", is_required=True
            ),
            GenerativeProviderCredentialConfig(env_var_name="AWS_SESSION_TOKEN", is_required=False),
        ],
    }

    @strawberry.field
    async def dependencies(self) -> list[str]:
        from phoenix.server.api.helpers.playground_registry import (
            PLAYGROUND_CLIENT_REGISTRY,
            PROVIDER_DEFAULT,
        )

        default_client = PLAYGROUND_CLIENT_REGISTRY.get_client(self.key, PROVIDER_DEFAULT)
        if default_client:
            return [dependency.name for dependency in default_client.dependencies()]
        return []

    @strawberry.field
    async def dependencies_installed(self) -> bool:
        from phoenix.server.api.helpers.playground_registry import (
            PLAYGROUND_CLIENT_REGISTRY,
            PROVIDER_DEFAULT,
        )

        default_client = PLAYGROUND_CLIENT_REGISTRY.get_client(self.key, PROVIDER_DEFAULT)
        if default_client:
            return default_client.dependencies_are_installed()
        return False

    @strawberry.field(description="The credential requirements for the provider")  # type: ignore
    async def credential_requirements(self) -> list[GenerativeProviderCredentialConfig]:
        # Handle providers that don't require credentials
        credential_requirements = self.model_provider_to_credential_requirements_map.get(self.key)
        if credential_requirements is None:
            return []
        return self.model_provider_to_credential_requirements_map[self.key]

    @strawberry.field(description="Whether the credentials are set on the server for the provider")  # type: ignore
    async def credentials_set(self) -> bool:
        # Check if every required credential is set
        return all(
            getenv(credential_config.env_var_name) is not None or not credential_config.is_required
            for credential_config in await self.credential_requirements()
        )

    @classmethod
    def _infer_model_provider_from_model_name(
        cls,
        model_name: str,
    ) -> Union[GenerativeProviderKey, None]:
        for provider, prefixes in cls.model_provider_to_model_prefix_map.items():
            if any(prefix.lower() in model_name.lower() for prefix in prefixes):
                return provider
        return None

    @classmethod
    def get_model_provider_from_attributes(
        cls,
        attributes: dict[str, Any],
    ) -> Union[GenerativeProviderKey, None]:
        llm_provider: Optional[str] = get_attribute_value(attributes, SpanAttributes.LLM_PROVIDER)

        if isinstance(llm_provider, str) and (
            provider := cls.attribute_provider_to_generative_provider_map.get(llm_provider)
        ):
            return provider
        llm_model = get_attribute_value(attributes, SpanAttributes.LLM_MODEL_NAME)
        if isinstance(llm_model, str):
            return cls._infer_model_provider_from_model_name(llm_model)
        return None



from datetime import datetime
from typing import TYPE_CHECKING, Optional

import strawberry
from strawberry.relay import Node, NodeID
from strawberry.types import Info
from typing_extensions import Annotated

from phoenix.db.models import ApiKey as OrmApiKey
from phoenix.server.api.context import Context

from .ApiKey import ApiKey

if TYPE_CHECKING:
    from .User import User


@strawberry.type
class UserApiKey(ApiKey, Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[OrmApiKey]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("UserApiKey ID mismatch")

    @strawberry.field(description="Name of the API key.")  # type: ignore
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.name),
            )
        return val

    @strawberry.field(description="Description of the API key.")  # type: ignore
    async def description(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.description
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.description),
            )
        return val

    @strawberry.field(description="The date and time the API key was created.")  # type: ignore
    async def created_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.created_at
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.created_at),
            )
        return val

    @strawberry.field(description="The date and time the API key will expire.")  # type: ignore
    async def expires_at(
        self,
        info: Info[Context, None],
    ) -> Optional[datetime]:
        if self.db_record:
            val = self.db_record.expires_at
        else:
            val = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.expires_at),
            )
        return val

    @strawberry.field
    async def user(
        self,
        info: Info[Context, None],
    ) -> Annotated["User", strawberry.lazy(".User")]:
        if self.db_record:
            user_id = self.db_record.user_id
        else:
            user_id = await info.context.data_loaders.user_api_key_fields.load(
                (self.id, OrmApiKey.user_id),
            )
        from .User import User

        return User(id=user_id)



# file: PromptLabelMutations.py

from typing import Optional

import strawberry
from sqlalchemy import delete
from sqlalchemy.exc import IntegrityError as PostgreSQLIntegrityError
from sqlean.dbapi2 import IntegrityError as SQLiteIntegrityError  # type: ignore[import-untyped]
from strawberry.relay import GlobalID
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.auth import IsLocked, IsNotReadOnly, IsNotViewer
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import Conflict, NotFound
from phoenix.server.api.queries import Query
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.api.types.Prompt import Prompt
from phoenix.server.api.types.PromptLabel import PromptLabel


@strawberry.input
class CreatePromptLabelInput:
    name: str
    description: Optional[str] = None
    color: str


@strawberry.input
class PatchPromptLabelInput:
    prompt_label_id: GlobalID
    name: Optional[str] = None
    description: Optional[str] = None


@strawberry.input
class DeletePromptLabelsInput:
    prompt_label_ids: list[GlobalID]


@strawberry.input
class SetPromptLabelsInput:
    prompt_id: GlobalID
    prompt_label_ids: list[GlobalID]


@strawberry.input
class UnsetPromptLabelsInput:
    prompt_id: GlobalID
    prompt_label_ids: list[GlobalID]


@strawberry.type
class PromptLabelMutationPayload:
    prompt_labels: list["PromptLabel"]
    query: "Query"


@strawberry.type
class PromptLabelDeleteMutationPayload:
    deleted_prompt_label_ids: list["GlobalID"]
    query: "Query"


@strawberry.type
class PromptLabelAssociationMutationPayload:
    query: "Query"


@strawberry.type
class PromptLabelMutationMixin:
    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer, IsLocked])  # type: ignore
    async def create_prompt_label(
        self, info: Info[Context, None], input: CreatePromptLabelInput
    ) -> PromptLabelMutationPayload:
        async with info.context.db() as session:
            label_orm = models.PromptLabel(
                name=input.name, description=input.description, color=input.color
            )
            session.add(label_orm)

            try:
                await session.commit()
            except (PostgreSQLIntegrityError, SQLiteIntegrityError):
                raise Conflict(f"A prompt label named '{input.name}' already exists.")

            return PromptLabelMutationPayload(
                prompt_labels=[PromptLabel(id=label_orm.id, db_record=label_orm)],
                query=Query(),
            )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer, IsLocked])  # type: ignore
    async def patch_prompt_label(
        self, info: Info[Context, None], input: PatchPromptLabelInput
    ) -> PromptLabelMutationPayload:
        async with info.context.db() as session:
            label_id = from_global_id_with_expected_type(
                input.prompt_label_id, PromptLabel.__name__
            )

            label_orm = await session.get(models.PromptLabel, label_id)
            if not label_orm:
                raise NotFound(f"PromptLabel with ID {input.prompt_label_id} not found")

            if input.name is not None:
                label_orm.name = input.name
            if input.description is not None:
                label_orm.description = input.description

            try:
                await session.commit()
            except (PostgreSQLIntegrityError, SQLiteIntegrityError):
                raise Conflict("Error patching PromptLabel. Possibly a name conflict?")

            return PromptLabelMutationPayload(
                prompt_labels=[PromptLabel(id=label_orm.id, db_record=label_orm)],
                query=Query(),
            )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def delete_prompt_labels(
        self, info: Info[Context, None], input: DeletePromptLabelsInput
    ) -> PromptLabelDeleteMutationPayload:
        """
        Deletes a PromptLabel (and any crosswalk references).
        """
        async with info.context.db() as session:
            label_ids = [
                from_global_id_with_expected_type(prompt_label_id, PromptLabel.__name__)
                for prompt_label_id in input.prompt_label_ids
            ]
            stmt = delete(models.PromptLabel).where(models.PromptLabel.id.in_(label_ids))
            await session.execute(stmt)

            await session.commit()

            return PromptLabelDeleteMutationPayload(
                deleted_prompt_label_ids=input.prompt_label_ids,
                query=Query(),
            )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer, IsLocked])  # type: ignore
    async def set_prompt_labels(
        self, info: Info[Context, None], input: SetPromptLabelsInput
    ) -> PromptLabelAssociationMutationPayload:
        async with info.context.db() as session:
            prompt_id = from_global_id_with_expected_type(input.prompt_id, Prompt.__name__)
            label_ids = [
                from_global_id_with_expected_type(prompt_label_id, PromptLabel.__name__)
                for prompt_label_id in input.prompt_label_ids
            ]

            crosswalk_items = [
                models.PromptPromptLabel(prompt_id=prompt_id, prompt_label_id=label_id)
                for label_id in label_ids
            ]
            session.add_all(crosswalk_items)

            try:
                await session.commit()
            except (PostgreSQLIntegrityError, SQLiteIntegrityError) as e:
                # The error could be:
                # - Unique constraint violation => row already exists
                # - Foreign key violation => prompt_id or label_id doesn't exist
                raise Conflict("Failed to associate PromptLabel with Prompt.") from e

            return PromptLabelAssociationMutationPayload(
                query=Query(),
            )

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def unset_prompt_labels(
        self, info: Info[Context, None], input: UnsetPromptLabelsInput
    ) -> PromptLabelAssociationMutationPayload:
        """
        Unsets a PromptLabel from a Prompt by removing the row in the crosswalk.
        """
        async with info.context.db() as session:
            prompt_id = from_global_id_with_expected_type(input.prompt_id, Prompt.__name__)
            label_ids = [
                from_global_id_with_expected_type(prompt_label_id, PromptLabel.__name__)
                for prompt_label_id in input.prompt_label_ids
            ]

            stmt = delete(models.PromptPromptLabel).where(
                (models.PromptPromptLabel.prompt_id == prompt_id)
                & (models.PromptPromptLabel.prompt_label_id.in_(label_ids))
            )
            result = await session.execute(stmt)

            if result.rowcount != len(label_ids):  # type: ignore[attr-defined]
                label_ids_str = ", ".join(str(i) for i in label_ids)
                raise NotFound(
                    f"No association between prompt={prompt_id} and labels={label_ids_str}."
                )

            await session.commit()

            return PromptLabelAssociationMutationPayload(
                query=Query(),
            )



from enum import Enum

import strawberry

from phoenix.core.model_schema import CONTINUOUS, Dimension


@strawberry.enum
class DimensionDataType(Enum):
    categorical = "categorical"
    numeric = "numeric"

    @classmethod
    def from_dimension(cls, dimension: Dimension) -> "DimensionDataType":
        data_type = dimension.data_type
        if data_type in (CONTINUOUS,):
            return cls.numeric
        return cls.categorical



from datetime import datetime
from typing import TYPE_CHECKING, Annotated, Optional

import strawberry
from sqlalchemy import func, select
from strawberry import UNSET, Private
from strawberry.relay import Connection, GlobalID, Node, NodeID
from strawberry.scalars import JSON
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.input_types.ExperimentRunSort import (
    ExperimentRunSort,
    add_order_by_and_page_start_to_query,
    get_experiment_run_cursor,
)
from phoenix.server.api.types.CostBreakdown import CostBreakdown
from phoenix.server.api.types.DatasetSplit import DatasetSplit
from phoenix.server.api.types.DatasetVersion import DatasetVersion
from phoenix.server.api.types.ExperimentAnnotationSummary import ExperimentAnnotationSummary
from phoenix.server.api.types.ExperimentRun import ExperimentRun
from phoenix.server.api.types.pagination import (
    ConnectionArgs,
    Cursor,
    CursorString,
    connection_from_cursors_and_nodes,
    connection_from_list,
)
from phoenix.server.api.types.SpanCostDetailSummaryEntry import SpanCostDetailSummaryEntry
from phoenix.server.api.types.SpanCostSummary import SpanCostSummary

_DEFAULT_EXPERIMENT_RUNS_PAGE_SIZE = 50

if TYPE_CHECKING:
    from .Project import Project


@strawberry.type
class Experiment(Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.Experiment]] = None
    cached_sequence_number: Private[Optional[int]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("Experiment ID mismatch")

    @strawberry.field
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.name),
            )
        return val

    @strawberry.field
    async def project_name(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.project_name
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.project_name),
            )
        return val

    @strawberry.field
    async def description(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.description
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.description),
            )
        return val

    @strawberry.field
    async def repetitions(
        self,
        info: Info[Context, None],
    ) -> int:
        if self.db_record:
            val = self.db_record.repetitions
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.repetitions),
            )
        return val

    @strawberry.field
    async def dataset_version_id(
        self,
        info: Info[Context, None],
    ) -> GlobalID:
        if self.db_record:
            version_id = self.db_record.dataset_version_id
        else:
            version_id = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.dataset_version_id),
            )
        return GlobalID(DatasetVersion.__name__, str(version_id))

    @strawberry.field
    async def metadata(
        self,
        info: Info[Context, None],
    ) -> JSON:
        if self.db_record:
            val = self.db_record.metadata_
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.metadata_),
            )
        return val

    @strawberry.field
    async def created_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.created_at
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.created_at),
            )
        return val

    @strawberry.field
    async def updated_at(
        self,
        info: Info[Context, None],
    ) -> datetime:
        if self.db_record:
            val = self.db_record.updated_at
        else:
            val = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.updated_at),
            )
        return val

    @strawberry.field(
        description="Sequence number (1-based) of experiments belonging to the same dataset"
    )  # type: ignore
    async def sequence_number(
        self,
        info: Info[Context, None],
    ) -> int:
        if self.cached_sequence_number is None:
            seq_num = await info.context.data_loaders.experiment_sequence_number.load(self.id)
            if seq_num is None:
                raise ValueError(f"invalid experiment: id={self.id}")
            self.cached_sequence_number = seq_num
        return self.cached_sequence_number

    @strawberry.field
    async def runs(
        self,
        info: Info[Context, None],
        first: Optional[int] = _DEFAULT_EXPERIMENT_RUNS_PAGE_SIZE,
        after: Optional[CursorString] = UNSET,
        sort: Optional[ExperimentRunSort] = UNSET,
    ) -> Connection[ExperimentRun]:
        if first is not None and first <= 0:
            raise BadRequest("first must be a positive integer if set")
        page_size = first or _DEFAULT_EXPERIMENT_RUNS_PAGE_SIZE
        experiment_runs_query = (
            select(models.ExperimentRun)
            .where(models.ExperimentRun.experiment_id == self.id)
            .limit(page_size + 1)
        )

        after_experiment_run_rowid = None
        after_sort_column_value = None
        if after:
            cursor = Cursor.from_string(after)
            after_experiment_run_rowid = cursor.rowid
            if cursor.sort_column is not None:
                after_sort_column_value = cursor.sort_column.value

        experiment_runs_query = add_order_by_and_page_start_to_query(
            query=experiment_runs_query,
            sort=sort,
            experiment_rowid=self.id,
            after_experiment_run_rowid=after_experiment_run_rowid,
            after_sort_column_value=after_sort_column_value,
        )

        async with info.context.db() as session:
            results = (await session.execute(experiment_runs_query)).all()

        has_next_page = False
        if len(results) > page_size:
            results = results[:page_size]
            has_next_page = True

        cursors_and_nodes = []
        for result in results:
            run = result[0]
            annotation_score = result[1] if len(result) > 1 else None
            gql_run = ExperimentRun(id=run.id, db_record=run)
            cursor = get_experiment_run_cursor(
                run=run, annotation_score=annotation_score, sort=sort
            )
            cursors_and_nodes.append((cursor, gql_run))

        return connection_from_cursors_and_nodes(
            cursors_and_nodes=cursors_and_nodes,
            has_previous_page=False,  # set to false since we are only doing forward pagination (https://relay.dev/graphql/connections.htm#sec-undefined.PageInfo.Fields) # noqa: E501
            has_next_page=has_next_page,
        )

    @strawberry.field
    async def run_count(self, info: Info[Context, None]) -> int:
        return await info.context.data_loaders.experiment_run_counts.load(self.id)

    @strawberry.field
    async def annotation_summaries(
        self, info: Info[Context, None]
    ) -> list[ExperimentAnnotationSummary]:
        experiment_id = self.id
        return [
            ExperimentAnnotationSummary(
                annotation_name=summary.annotation_name,
                min_score=summary.min_score,
                max_score=summary.max_score,
                mean_score=summary.mean_score,
                count=summary.count,
                error_count=summary.error_count,
            )
            for summary in await info.context.data_loaders.experiment_annotation_summaries.load(
                experiment_id
            )
        ]

    @strawberry.field
    async def error_rate(self, info: Info[Context, None]) -> Optional[float]:
        return await info.context.data_loaders.experiment_error_rates.load(self.id)

    @strawberry.field
    async def average_run_latency_ms(self, info: Info[Context, None]) -> Optional[float]:
        latency_ms = await info.context.data_loaders.average_experiment_run_latency.load(self.id)
        return latency_ms

    @strawberry.field
    async def project(
        self, info: Info[Context, None]
    ) -> Optional[Annotated["Project", strawberry.lazy(".Project")]]:
        if self.db_record:
            project_name = self.db_record.project_name
        else:
            project_name = await info.context.data_loaders.experiment_fields.load(
                (self.id, models.Experiment.project_name),
            )

        if project_name is None:
            return None

        db_project = await info.context.data_loaders.project_by_name.load(project_name)

        if db_project is None:
            return None
        from .Project import Project

        return Project(id=db_project.id, db_record=db_project)

    @strawberry.field
    def last_updated_at(self, info: Info[Context, None]) -> Optional[datetime]:
        return info.context.last_updated_at.get(models.Experiment, self.id)

    @strawberry.field
    async def cost_summary(self, info: Info[Context, None]) -> SpanCostSummary:
        summary = await info.context.data_loaders.span_cost_summary_by_experiment.load(self.id)
        return SpanCostSummary(
            prompt=CostBreakdown(
                tokens=summary.prompt.tokens,
                cost=summary.prompt.cost,
            ),
            completion=CostBreakdown(
                tokens=summary.completion.tokens,
                cost=summary.completion.cost,
            ),
            total=CostBreakdown(
                tokens=summary.total.tokens,
                cost=summary.total.cost,
            ),
        )

    @strawberry.field
    async def cost_detail_summary_entries(
        self, info: Info[Context, None]
    ) -> list[SpanCostDetailSummaryEntry]:
        stmt = (
            select(
                models.SpanCostDetail.token_type,
                models.SpanCostDetail.is_prompt,
                func.sum(models.SpanCostDetail.cost).label("cost"),
                func.sum(models.SpanCostDetail.tokens).label("tokens"),
            )
            .select_from(models.SpanCostDetail)
            .join(models.SpanCost, models.SpanCostDetail.span_cost_id == models.SpanCost.id)
            .join(models.Span, models.SpanCost.span_rowid == models.Span.id)
            .join(models.Trace, models.Span.trace_rowid == models.Trace.id)
            .join(models.ExperimentRun, models.ExperimentRun.trace_id == models.Trace.trace_id)
            .where(models.ExperimentRun.experiment_id == self.id)
            .group_by(models.SpanCostDetail.token_type, models.SpanCostDetail.is_prompt)
        )

        async with info.context.db() as session:
            data = await session.stream(stmt)
            return [
                SpanCostDetailSummaryEntry(
                    token_type=token_type,
                    is_prompt=is_prompt,
                    value=CostBreakdown(tokens=tokens, cost=cost),
                )
                async for token_type, is_prompt, cost, tokens in data
            ]

    @strawberry.field
    async def dataset_splits(
        self,
        info: Info[Context, None],
    ) -> Connection[DatasetSplit]:
        """Returns the dataset splits associated with this experiment."""
        splits = await info.context.data_loaders.experiment_dataset_splits.load(self.id)
        return connection_from_list(
            [DatasetSplit(id=split.id, db_record=split) for split in splits], ConnectionArgs()
        )


def to_gql_experiment(
    experiment: models.Experiment,
    sequence_number: Optional[int] = None,
) -> Experiment:
    """
    Converts an ORM experiment to a GraphQL Experiment.
    """
    return Experiment(
        id=experiment.id,
        db_record=experiment,
        cached_sequence_number=sequence_number,
    )



from math import isfinite
from typing import TYPE_CHECKING, Annotated, Optional

import strawberry
from strawberry.relay import Node, NodeID
from strawberry.scalars import JSON
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.context import Context
from phoenix.server.api.types.AnnotatorKind import AnnotatorKind

from .AnnotationSource import AnnotationSource

if TYPE_CHECKING:
    from .Trace import Trace
    from .User import User


@strawberry.type
class TraceAnnotation(Node):
    id: NodeID[int]
    db_record: strawberry.Private[Optional[models.TraceAnnotation]] = None

    def __post_init__(self) -> None:
        if self.db_record and self.id != self.db_record.id:
            raise ValueError("TraceAnnotation ID mismatch")

    @strawberry.field(description="Name of the annotation, e.g. 'helpfulness' or 'relevance'.")  # type: ignore
    async def name(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.name
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.name),
            )
        return val

    @strawberry.field(description="The kind of annotator that produced the annotation.")  # type: ignore
    async def annotator_kind(
        self,
        info: Info[Context, None],
    ) -> AnnotatorKind:
        if self.db_record:
            val = self.db_record.annotator_kind
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.annotator_kind),
            )
        return AnnotatorKind(val)

    @strawberry.field(
        description="Value of the annotation in the form of a string, e.g. 'helpful' or 'not helpful'. Note that the label is not necessarily binary."  # noqa: E501
    )  # type: ignore
    async def label(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.label
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.label),
            )
        return val

    @strawberry.field(description="Value of the annotation in the form of a numeric score.")  # type: ignore
    async def score(
        self,
        info: Info[Context, None],
    ) -> Optional[float]:
        if self.db_record:
            val = self.db_record.score
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.score),
            )
        return val if val is not None and isfinite(val) else None

    @strawberry.field(
        description="The annotator's explanation for the annotation result (i.e. score or label, or both) given to the subject."  # noqa: E501
    )  # type: ignore
    async def explanation(
        self,
        info: Info[Context, None],
    ) -> Optional[str]:
        if self.db_record:
            val = self.db_record.explanation
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.explanation),
            )
        return val

    @strawberry.field(description="Metadata about the annotation.")  # type: ignore
    async def metadata(
        self,
        info: Info[Context, None],
    ) -> JSON:
        if self.db_record:
            val = self.db_record.metadata_
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.metadata_),
            )
        return val

    @strawberry.field(description="The identifier of the annotation.")  # type: ignore
    async def identifier(
        self,
        info: Info[Context, None],
    ) -> str:
        if self.db_record:
            val = self.db_record.identifier
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.identifier),
            )
        return val

    @strawberry.field(description="The source of the annotation.")  # type: ignore
    async def source(
        self,
        info: Info[Context, None],
    ) -> AnnotationSource:
        if self.db_record:
            val = self.db_record.source
        else:
            val = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.source),
            )
        return AnnotationSource(val)

    @strawberry.field(description="The trace associated with the annotation.")  # type: ignore
    async def trace(
        self,
        info: Info[Context, None],
    ) -> Annotated["Trace", strawberry.lazy(".Trace")]:
        if self.db_record:
            trace_rowid = self.db_record.trace_rowid
        else:
            trace_rowid = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.trace_rowid),
            )
        from .Trace import Trace

        return Trace(id=trace_rowid)

    @strawberry.field(description="The user that produced the annotation.")  # type: ignore
    async def user(
        self,
        info: Info[Context, None],
    ) -> Optional[Annotated["User", strawberry.lazy(".User")]]:
        if self.db_record:
            user_id = self.db_record.user_id
        else:
            user_id = await info.context.data_loaders.trace_annotation_fields.load(
                (self.id, models.TraceAnnotation.user_id),
            )
        if user_id is None:
            return None
        from .User import User

        return User(id=user_id)



from dataclasses import dataclass

from sqlalchemy import select, tuple_
from strawberry.dataloader import DataLoader
from typing_extensions import TypeAlias

from phoenix.db import models
from phoenix.server.types import DbSessionFactory

ExperimentID: TypeAlias = int
DatasetExampleID: TypeAlias = int
Key: TypeAlias = tuple[ExperimentID, DatasetExampleID]


@dataclass
class ExperimentRepeatedRunGroup:
    experiment_rowid: int
    dataset_example_rowid: int
    runs: list[models.ExperimentRun]


Result: TypeAlias = ExperimentRepeatedRunGroup


class ExperimentRepeatedRunGroupsDataLoader(DataLoader[Key, Result]):
    def __init__(self, db: DbSessionFactory) -> None:
        super().__init__(load_fn=self._load_fn)
        self._db = db

    async def _load_fn(self, keys: list[Key]) -> list[Result]:
        repeated_run_groups_query = (
            select(models.ExperimentRun)
            .where(
                tuple_(
                    models.ExperimentRun.experiment_id,
                    models.ExperimentRun.dataset_example_id,
                ).in_(set(keys))
            )
            .order_by(models.ExperimentRun.repetition_number)
        )

        async with self._db() as session:
            runs_by_key: dict[Key, list[models.ExperimentRun]] = {}
            for run in (await session.scalars(repeated_run_groups_query)).all():
                key = (run.experiment_id, run.dataset_example_id)
                if key not in runs_by_key:
                    runs_by_key[key] = []
                runs_by_key[key].append(run)

        return [
            ExperimentRepeatedRunGroup(
                experiment_rowid=experiment_id,
                dataset_example_rowid=dataset_example_id,
                runs=runs_by_key.get((experiment_id, dataset_example_id), []),
            )
            for (experiment_id, dataset_example_id) in keys
        ]



from collections.abc import Iterable

from openinference.semconv.resource import ResourceAttributes
from opentelemetry.proto.common.v1.common_pb2 import KeyValue

from phoenix.config import DEFAULT_PROJECT_NAME


def get_project_name(attributes: Iterable[KeyValue]) -> str:
    for kv in attributes:
        if kv.key == ResourceAttributes.PROJECT_NAME and (v := kv.value.string_value):
            return str(v)
    return DEFAULT_PROJECT_NAME



from typing import Any

from fastapi.openapi.utils import get_openapi

from phoenix.server.api.routers.v1 import REST_API_VERSION, create_v1_router


def get_openapi_schema() -> dict[str, Any]:
    v1_router = create_v1_router(authentication_enabled=False)
    return get_openapi(
        title="Arize-Phoenix REST API",
        version=REST_API_VERSION,
        openapi_version="3.1.0",
        description="Schema for Arize-Phoenix REST API",
        routes=v1_router.routes,
        separate_input_output_schemas=False,
    )



from datetime import datetime
from typing import Optional, Sequence

from openinference.semconv.trace import SpanAttributes
from sqlalchemy import distinct, or_, select
from sqlalchemy.sql.selectable import ScalarSelect

from phoenix.db import models


def get_filtered_session_rowids_subquery(
    session_filter_condition: str,
    project_rowids: Sequence[int],
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
) -> ScalarSelect[int]:
    """
    Returns a subquery that contains the project session rowids that match the session filter.
    """

    filtered_session_rowids = (
        select(distinct(models.Trace.project_session_rowid).label("id"))
        .join_from(models.Trace, models.Span)
        .where(models.Trace.project_rowid.in_(project_rowids))
        .where(models.Span.parent_id.is_(None))
        .where(
            or_(
                models.CaseInsensitiveContains(
                    models.Span.attributes[INPUT_VALUE].as_string(),
                    session_filter_condition,
                ),
                models.CaseInsensitiveContains(
                    models.Span.attributes[OUTPUT_VALUE].as_string(),
                    session_filter_condition,
                ),
            )
        )
    )
    if start_time:
        filtered_session_rowids = filtered_session_rowids.where(
            start_time <= models.Trace.start_time
        )
    if end_time:
        filtered_session_rowids = filtered_session_rowids.where(models.Trace.start_time < end_time)
    return filtered_session_rowids.scalar_subquery()


INPUT_VALUE = SpanAttributes.INPUT_VALUE.split(".")
OUTPUT_VALUE = SpanAttributes.OUTPUT_VALUE.split(".")



from __future__ import annotations

import ssl
from typing import Any, Container, Final, Literal, Mapping, TypedDict, get_type_hints

from sqlalchemy import URL
from typing_extensions import assert_never


def get_pg_config(
    url: URL,
    driver: Literal["psycopg", "asyncpg"],
    enforce_ssl: bool = False,
) -> tuple[URL, dict[str, Any]]:
    """Convert SQLAlchemy URL to driver-specific configuration.

    Args:
        url: SQLAlchemy URL
        driver: "psycopg" or "asyncpg"
        enforce_ssl: If True, ensure SSL is enabled (required for AWS RDS IAM auth)

    Returns:
        Tuple of (base_url, connect_args):
        - base_url: URL with driver prefix and non-SSL parameters
        - connect_args: SSL configuration for the driver
    """
    # Create new URL with appropriate driver
    query = url.query
    ssl_args = _get_ssl_args(query)

    if enforce_ssl and not ssl_args:
        ssl_args = {"sslmode": "require"}
    elif enforce_ssl and ssl_args.get("sslmode") == "disable":
        raise ValueError(
            "SSL cannot be disabled when using AWS RDS IAM authentication. "
            "Remove 'sslmode=disable' from the connection string."
        )

    # Create base URL without SSL parameters
    base_url = url.set(
        drivername=f"postgresql+{driver}",
        query={k: v for k, v in query.items() if k not in _SSL_KEYS},
    )

    # Get appropriate SSL configuration based on driver
    if driver == "psycopg":
        connect_args = dict(ssl_args)
        # Remove asyncpg-specific parameters from base URL
        base_url = base_url.set(query=_remove_asyncpg_only_params(base_url.query))
    elif driver == "asyncpg":
        # Only create SSL context if we have SSL parameters and sslmode is not disable
        if ssl_args and ssl_args.get("sslmode") != "disable":
            connect_args = {"ssl": _get_ssl_context(ssl_args)}
        else:
            connect_args = {}
    else:
        assert_never(driver)
    return base_url, connect_args


class _SSLArgs(TypedDict, total=False):
    """SSL parameters for PostgreSQL connections.

    All fields are optional. Only includes parameters that can be converted to ssl.SSLContext.

    Attributes:
        sslmode: SSL mode (disable, allow, prefer, require, verify-ca, verify-full)
        sslrootcert: Path to root CA certificate
        sslcert: Path to client certificate
        sslkey: Path to client private key
        sslpassword: Password for private key
        sslcrl: Path to CRL file
        sslcrldir: Path to CRL directory
        sslsni: Enable SNI (0 or 1)
    """

    sslmode: str
    sslrootcert: str
    sslcert: str
    sslkey: str
    sslpassword: str
    sslcrl: str
    sslcrldir: str
    sslsni: str


_SSL_KEYS: Final[Container[str]] = get_type_hints(_SSLArgs).keys()


def _get_ssl_args(
    query_params: Mapping[str, str | tuple[str, ...]],
) -> _SSLArgs:
    """Extract SSL parameters from a SQLAlchemy URL query.

    Args:
        query_params: SQLAlchemy URL query parameters

    Returns:
        Dictionary of SSL parameters
    """
    result: _SSLArgs = {}

    def get_str(key: str) -> str | None:
        if value := query_params.get(key):
            if not isinstance(value, str):
                raise ValueError(f"Invalid value type for {key}: {type(value)}")
            return value
        return None

    if sslmode := get_str("sslmode"):
        result["sslmode"] = sslmode

    if ca_cert := get_str("sslrootcert"):
        result["sslrootcert"] = ca_cert

    if cert := get_str("sslcert"):
        result["sslcert"] = cert

    if key := get_str("sslkey"):
        result["sslkey"] = key

    if password := get_str("sslpassword"):
        result["sslpassword"] = password

    if crl := get_str("sslcrl"):
        result["sslcrl"] = crl

    if crl_dir := get_str("sslcrldir"):
        result["sslcrldir"] = crl_dir

    if sslsni := get_str("sslsni"):
        result["sslsni"] = sslsni

    return result


def _get_ssl_context(ssl_args: _SSLArgs) -> ssl.SSLContext:
    """Convert PostgreSQL SSL parameters to an SSL context.

    Args:
        ssl_args: PostgreSQL SSL parameters from _SSLArgs TypedDict

    Returns:
        Configured SSL context with:
        - Root CA certificate for server verification (if provided)
        - Client certificate and key for mutual TLS (if provided)
        - SSL mode appropriate verification settings
        - Certificate revocation list checking (if provided)
    """
    # Create SSL context
    ssl_context = ssl.create_default_context()

    # Load CA certificate if provided
    if ca_cert := ssl_args.get("sslrootcert"):
        ssl_context.load_verify_locations(cafile=ca_cert)

    # Load client certificates if provided
    if (cert := ssl_args.get("sslcert")) and (key := ssl_args.get("sslkey")):
        ssl_context.load_cert_chain(
            certfile=cert,
            keyfile=key,
            password=ssl_args.get("sslpassword"),
        )

    # Load CRL if provided
    if crl := ssl_args.get("sslcrl"):
        ssl_context.load_verify_locations(cafile=crl)
    if crl_dir := ssl_args.get("sslcrldir"):
        ssl_context.load_verify_locations(capath=crl_dir)

    # Set verification mode based on sslmode
    sslmode = ssl_args.get("sslmode", "prefer")
    if sslmode == "verify-full":
        # Full verification: certificate and hostname
        ssl_context.check_hostname = True
        ssl_context.verify_mode = ssl.CERT_REQUIRED
    elif sslmode == "verify-ca":
        # Certificate verification only
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_REQUIRED
    else:  # require, prefer, allow, disable
        # No verification, just encryption
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE

    return ssl_context


def _remove_asyncpg_only_params(
    query: Mapping[str, str | tuple[str, ...]],
) -> dict[str, str | tuple[str, ...]]:
    """Remove asyncpg-specific parameters from a SQLAlchemy URL query.

    Args:
        query: SQLAlchemy URL query parameters

    Returns:
        Dictionary of query parameters with asyncpg-specific parameters removed
    """
    return {k: v for k, v in query.items() if k not in _ASYNCPG_ONLY_KEYS}


# Asyncpg-specific parameter keys
_ASYNCPG_ONLY_KEYS: Final[tuple[str, ...]] = (
    "prepared_statement_cache_size",
    # Add other asyncpg-specific parameters here if needed
)



import strawberry
from sqlalchemy import and_, delete, not_, select, update
from sqlalchemy.orm import load_only
from sqlalchemy.sql import literal
from strawberry.relay import GlobalID
from strawberry.types import Info

from phoenix.db import models
from phoenix.server.api.auth import IsNotReadOnly, IsNotViewer
from phoenix.server.api.context import Context
from phoenix.server.api.exceptions import BadRequest
from phoenix.server.api.queries import Query
from phoenix.server.api.types.node import from_global_id_with_expected_type
from phoenix.server.dml_event import SpanDeleteEvent


@strawberry.type
class TraceMutationMixin:
    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def delete_traces(
        self,
        info: Info[Context, None],
        trace_ids: list[GlobalID],
    ) -> Query:
        if not trace_ids:
            raise BadRequest("Must provide at least one trace ID to delete")
        trace_ids = list(set(trace_ids))
        try:
            trace_rowids = [
                from_global_id_with_expected_type(global_id=id, expected_type_name="Trace")
                for id in trace_ids
            ]
        except ValueError as error:
            raise BadRequest(str(error))
        async with info.context.db() as session:
            traces = (
                await session.scalars(
                    delete(models.Trace)
                    .where(models.Trace.id.in_(trace_rowids))
                    .returning(models.Trace)
                    .options(
                        load_only(models.Trace.project_rowid, models.Trace.project_session_rowid)
                    )
                )
            ).all()
            if len(traces) < len(trace_rowids):
                await session.rollback()
                raise BadRequest("Invalid trace IDs provided")
            project_ids = tuple(set(trace.project_rowid for trace in traces))
            if len(project_ids) > 1:
                await session.rollback()
                raise BadRequest("Cannot delete traces from multiple projects")
            session_ids = set(
                session_id
                for trace in traces
                if (session_id := trace.project_session_rowid) is not None
            )
            if session_ids:
                await session.execute(
                    delete(models.ProjectSession).where(
                        and_(
                            models.ProjectSession.id.in_(session_ids),
                            not_(
                                select(literal(1))
                                .where(
                                    models.Trace.project_session_rowid == models.ProjectSession.id
                                )
                                .exists()
                            ),
                        )
                    )
                )
            info.context.event_queue.put(SpanDeleteEvent(project_ids))
        return Query()

    @strawberry.mutation(permission_classes=[IsNotReadOnly, IsNotViewer])  # type: ignore
    async def transfer_traces_to_project(
        self,
        info: Info[Context, None],
        trace_ids: list[GlobalID],
        project_id: GlobalID,
    ) -> Query:
        if not trace_ids:
            raise BadRequest("Must provide at least one trace ID to transfer")
        trace_ids = list(set(trace_ids))
        try:
            trace_rowids = [
                from_global_id_with_expected_type(global_id=id, expected_type_name="Trace")
                for id in trace_ids
            ]
            dest_project_rowid = from_global_id_with_expected_type(
                global_id=project_id, expected_type_name="Project"
            )
        except ValueError as error:
            raise BadRequest(str(error))

        async with info.context.db() as session:
            dest_project = await session.get(models.Project, dest_project_rowid)
            if dest_project is None:
                raise BadRequest("Destination project does not exist")

            traces = (
                await session.scalars(select(models.Trace).where(models.Trace.id.in_(trace_rowids)))
            ).all()
            if len(traces) < len(trace_rowids):
                raise BadRequest("Invalid trace IDs provided")

            source_project_ids = set(trace.project_rowid for trace in traces)
            if len(source_project_ids) > 1:
                raise BadRequest("Cannot transfer traces from multiple projects")

            await session.execute(
                update(models.Trace)
                .where(models.Trace.id.in_(trace_rowids))
                .values(project_rowid=dest_project_rowid)
            )

        return Query()



import base64
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto
from typing import Any, ClassVar, Optional, Union

from strawberry import UNSET
from strawberry.relay.types import Connection, Edge, NodeType, PageInfo
from typing_extensions import TypeAlias, assert_never

ID: TypeAlias = int
CursorSortColumnValue: TypeAlias = Union[str, int, float, datetime, None]

# A type alias for the connection cursor implementation
CursorString = str

# The hashing prefix for a connection cursor
CURSOR_PREFIX = "connection:"


class CursorSortColumnDataType(Enum):
    NULL = auto()
    STRING = auto()
    INT = auto()
    FLOAT = auto()
    DATETIME = auto()


@dataclass
class CursorSortColumn:
    type: CursorSortColumnDataType
    value: CursorSortColumnValue

    def __post_init__(self) -> None:
        if self.value is None:
            self.type = CursorSortColumnDataType.NULL

    def __str__(self) -> str:
        if isinstance(self.value, str):
            return self.value
        if isinstance(self.value, (int, float)):
            return str(self.value)
        if isinstance(self.value, datetime):
            return self.value.isoformat()
        if self.value is None:
            return ""
        assert_never(self.value)

    @classmethod
    def from_string(cls, type: CursorSortColumnDataType, cursor_string: str) -> "CursorSortColumn":
        value: CursorSortColumnValue
        if type is CursorSortColumnDataType.STRING:
            value = cursor_string
        elif type is CursorSortColumnDataType.INT:
            value = int(cursor_string)
        elif type is CursorSortColumnDataType.FLOAT:
            value = float(cursor_string)
        elif type is CursorSortColumnDataType.DATETIME:
            value = datetime.fromisoformat(cursor_string)
        elif type is CursorSortColumnDataType.NULL:
            value = None
        else:
            assert_never(type)
        return cls(type=type, value=value)


@dataclass
class Cursor:
    """
    Serializes and deserializes cursor strings for ID-based pagination.

    In the simplest case, a cursor encodes the rowid of a record. In the case
    that a sort has been applied, the cursor additionally encodes the data type
    and value of the column indexed for sorting so that the sort position can be
    efficiently found. The encoding ensures that the cursor string is opaque to
    the client and discourages the client from making use of the encoded
    content.

    Examples:
        # encodes "10"
        Cursor(rowid=10)

        # encodes "11:STRING:abc"
        Cursor(
            rowid=11,
            sort_column=CursorSortColumn(
                type=CursorSortColumnDataType.STRING,
                value="abc"
            )
        )

        # encodes "10:INT:5"
        Cursor(
            rowid=10,
            sort_column=CursorSortColumn(
                type=CursorSortColumnDataType.INT,
                value=5
            )
        )

        # encodes "17:FLOAT:5.7"
        Cursor(
            rowid=17,
            sort_column=CursorSortColumn(
                type=CursorSortColumnDataType.FLOAT,
                value=5.7
            )
        )

        # encodes "20:DATETIME:2024-05-05T04:25:29.911245+00:00"
        Cursor(
            rowid=20,
            sort_column=CursorSortColumn(
                type=CursorSortColumnDataType.DATETIME,
                value=datetime.fromisoformat("2024-05-05T04:25:29.911245+00:00")
            )
        )
    """

    rowid: int
    sort_column: Optional[CursorSortColumn] = None

    _DELIMITER: ClassVar[str] = ":"

    def __str__(self) -> str:
        cursor_parts = [str(self.rowid)]
        if (sort_column := self.sort_column) is not None:
            cursor_parts.extend([sort_column.type.name, str(sort_column)])
        return base64.b64encode(self._DELIMITER.join(cursor_parts).encode()).decode()

    @classmethod
    def from_string(cls, cursor: str) -> "Cursor":
        decoded = base64.b64decode(cursor).decode()
        rowid_string = decoded
        sort_column = None
        if (first_delimiter_index := decoded.find(cls._DELIMITER)) > -1:
            rowid_string = decoded[:first_delimiter_index]
            second_delimiter_index = decoded.index(cls._DELIMITER, first_delimiter_index + 1)
            sort_column = CursorSortColumn.from_string(
                type=CursorSortColumnDataType[
                    decoded[first_delimiter_index + 1 : second_delimiter_index]
                ],
                cursor_string=decoded[second_delimiter_index + 1 :],
            )
        return cls(rowid=int(rowid_string), sort_column=sort_column)


def offset_to_cursor(offset: int) -> CursorString:
    """
    Creates the cursor string from an offset.
    """
    return base64.b64encode(f"{CURSOR_PREFIX}{offset}".encode("utf-8")).decode()


def cursor_to_offset(cursor: CursorString) -> int:
    """
    Extracts the offset from the cursor string.
    """
    prefix, offset = base64.b64decode(cursor).decode().split(":")
    return int(offset)


def get_offset_with_default(cursor: Optional[CursorString], default_offset: int) -> int:
    """
    Given an optional cursor and a default offset, returns the offset
    to use; if the cursor contains a valid offset, that will be used,
    otherwise it will be the default.
    """
    if not isinstance(cursor, CursorString):
        return default_offset
    offset = cursor_to_offset(cursor)
    return offset if isinstance(offset, int) else default_offset


@dataclass(frozen=True)
class ConnectionArgs:
    """
    Arguments common to all connections
    """

    first: Optional[int] = UNSET
    after: Optional[CursorString] = UNSET
    last: Optional[int] = UNSET
    before: Optional[CursorString] = UNSET


def connection_from_list(
    data: list[NodeType],
    args: ConnectionArgs,
) -> Connection[NodeType]:
    """
    A simple function that accepts a list and connection arguments, and returns
    a connection object for use in GraphQL. It uses list offsets as pagination,
    so pagination will only work if the list is static.
    """
    return connection_from_list_slice(data, args, slice_start=0, list_length=len(data))


def connection_from_list_slice(
    list_slice: list[NodeType],
    args: ConnectionArgs,
    slice_start: int,
    list_length: int,
) -> Connection[NodeType]:
    """
    Given a slice (subset) of a list, returns a connection object for use in
    GraphQL.

    This function is similar to `connection_from_list`, but is intended for use
    cases where you know the cardinality of the connection, consider it too large
    to materialize the entire list, and instead wish pass in a slice of the
    total result large enough to cover the range specified in `args`.
    """

    slice_end = slice_start + len(list_slice)

    start_offset = max(slice_start, 0)
    end_offset = min(slice_end, list_length)

    after_offset = get_offset_with_default(args.after, -1)

    if 0 <= after_offset < list_length:
        start_offset = max(start_offset, after_offset + 1)

    before_offset = get_offset_with_default(args.before, end_offset)

    if 0 <= before_offset < list_length:
        end_offset = min(end_offset, before_offset)

    if isinstance(args.first, int):
        if args.first < 0:
            raise Exception('Argument "first" must be a non-negative int')
        end_offset = min(end_offset, start_offset + args.first)

    if isinstance(args.last, int):
        if args.last < 0:
            raise Exception('Argument "last" must be a non-negative int')
        start_offset = max(start_offset, end_offset - args.last)

    # If supplied slice is too large, trim it down before mapping over it.
    slice = list_slice[start_offset - slice_start : end_offset - slice_start]

    edges = [
        Edge(node=node, cursor=offset_to_cursor(start_offset + index))
        for index, node in enumerate(slice)
    ]

    has_edges = len(edges) > 0
    first_edge = edges[0] if has_edges else None
    last_edge = edges[-1] if has_edges else None
    lower_bound = after_offset + 1 if args.after is not None else 0
    upper_bound = before_offset if args.before is not None else list_length

    return Connection(
        edges=edges,
        page_info=PageInfo(
            start_cursor=first_edge.cursor if first_edge else None,
            end_cursor=last_edge.cursor if last_edge else None,
            has_previous_page=start_offset > lower_bound if isinstance(args.last, int) else False,
            has_next_page=end_offset < upper_bound if isinstance(args.first, int) else False,
        ),
    )


def connection_from_cursors_and_nodes(
    cursors_and_nodes: list[tuple[Any, NodeType]],
    has_previous_page: bool,
    has_next_page: bool,
) -> Connection[NodeType]:
    edges = [Edge(node=node, cursor=str(cursor)) for cursor, node in cursors_and_nodes]
    has_edges = len(edges) > 0
    first_edge = edges[0] if has_edges else None
    last_edge = edges[-1] if has_edges else None
    return Connection(
        edges=edges,
        page_info=PageInfo(
            start_cursor=first_edge.cursor if first_edge else None,
            end_cursor=last_edge.cursor if last_edge else None,
            has_previous_page=has_previous_page,
            has_next_page=has_next_page,
        ),
    )
